"""
Prediction script crYOLO
"""

# ! /usr/bin/env python
#
# COPYRIGHT
#
# All contributions by Ngoc Anh Huyn:
# Copyright (c) 2017, Ngoc Anh Huyn.
# All rights reserved.
#
# All contributions by Thorsten Wagner:
# Copyright (c) 2017 - 2019, Thorsten Wagner.
# All rights reserved.
#
# ---------------------------------------------------------------------------
#         Do not reproduce or redistribute, in whole or in part.
#      Use of this code is permitted only under licence from Max Planck Society.
#            Contact us at thorsten.wagner@mpi-dortmund.mpg.de
# ---------------------------------------------------------------------------
#
from __future__ import print_function
import multiprocessing
import time
import argparse
import os
import sys
import json
import re
import numpy as np
from cryolo.utils import BoundBox
from lineenhancer import line_enhancer, maskstackcreator
from . import CoordsIO
from . import imagereader
from . import utils
from . import filament_tracer
from . import config_tools
from gooey import Gooey, GooeyParser
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler
from tqdm import tqdm

os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
try:
    os.environ["CUDA_VISIBLE_DEVICES"]
except KeyError:
    os.environ["CUDA_VISIBLE_DEVICES"] = "0"
os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"


ARGPARSER = None
MONITOR=False
filament_tracers = []


def create_parser(parser, use_gooey=True):
    """

    :param parser: Parser where the subgroups are added
    :param use_gooey: ture if gooey is used
    :return:
    """

    required_group = parser.add_argument_group(
        "Required arguments", "These options are mandatory to run crYOLO prediction"
    )
    args = ["-c", "--conf"]
    kwargs = {
        "help": "Path to the crYOLO configuration file.",
        "required": True,
        "gooey_options": {
            "validator": {
                "test": 'user_input.endswith("json")',
                "message": "File has to end with .json!",
            },
            "wildcard": "*.json",
        },
    }

    if use_gooey:
        kwargs["widget"] = "FileChooser"
    required_group.add_argument(*args, **kwargs)

    args = ["-w", "--weights"]
    kwargs = {
        "help": "Path to the trained model. It can either be a model that you trained from scratch, a refined model or a general model.",
        "required": True,
        "gooey_options": {
            "validator": {
                "test": 'user_input.endswith("h5")',
                "message": "File has to end with .h5!",
            },
            "wildcard": "*.h5",
        },
    }

    if use_gooey:
        kwargs["widget"] = "FileChooser"
    required_group.add_argument(*args, **kwargs)

    args = ["-i", "--input"]
    kwargs = {
        "nargs": "+",
        "help": "Path to one or multiple image folders / images (only directories in GUI).",
        "required": True,
    }
    if use_gooey:
        kwargs["widget"] = "DirChooser"
    required_group.add_argument(*args, **kwargs)

    args = ["-o", "--output"]
    kwargs = {
        "help": "Path to the output folder. All particle coordinates will be written there.",
        "required": True,
    }
    if use_gooey:
        kwargs["widget"] = "DirChooser"

    required_group.add_argument(*args, **kwargs)

    optional_group = parser.add_argument_group(
        "Optional arguments", "Optional arguments for crYOLO"
    )

    optional_group.add_argument(
        "-t",
        "--threshold",
        type=float,
        default=0.3,
        help="Confidence threshold. Have to be between 0 and 1. The higher, the more conservative.",
        gooey_options={
            "validator": {
                "test": "0.0 <= float(user_input) <= 1.0",
                "message": "Must be between 0.0 and 1.0",
            }
        },
    )

    optional_group.add_argument(
        "-g",
        "--gpu",
        default="",
        # type=int,
        nargs="+",
        help="Specify which gpu(s) should be used. Multiple GPUs are separated by a whitespace. If not defined otherwise by your system, it will use GPU 0 by default.",
    )

    optional_group.add_argument(
        "-d",
        "--distance",
        default=0,
        type=int,
        help="Particles with a distance less than this value (in pixel) will be removed. This option should not be used in filament mode.",
    )

    optional_group.add_argument(
        "--minsize",
        type=int,
        help="Particles with a estimated diameter less than this value (in pixel) will be removed. This option typically is only useful for the general model.",
    )

    optional_group.add_argument(
        "--maxsize",
        type=int,
        help="Particles with a estimated diameter greather than this value (in pixel) will be removed.  This option typically is only useful for the general model.",
    )

    optional_group.add_argument(
        "-pbs",
        "--prediction_batch_size",
        default=3,
        type=int,
        help="How many images should be predicted in one batch. Smaller values might resolve memory issues.",
    )

    optional_group.add_argument(
        "--gpu_fraction",
        type=float,
        default=1.0,
        help="Specify the fraction of memory per GPU used by crYOLO during prediction. Only values between 0.0 and 1.0 are allowed.",
        gooey_options={
            "validator": {
                "test": "0.0 <= float(user_input) <= 1.0",
                "message": "Must be between 0.0 and 1.0",
            }
        },
    )

    optional_group.add_argument(
        "-nc",
        "--num_cpu",
        type=int,
        default=-1,
        help="Number of CPUs used during filtering / filament tracing. By default it will use all of the available CPUs.",
    )

    optional_group.add_argument(
        "--norm_margin",
        type=float,
        default=0.0,
        help="Relative margin size for normalization.",
        gooey_options={
            "validator": {
                "test": "0.0 <= float(user_input) <= 1.0",
                "message": "Must be between 0.0 and 1.0",
            }
        },
    )

    optional_group.add_argument(
        "--monitor",
        action="store_true",
        help='When this option is activated, crYOLO will monitor your input folder. This especially useful for automation purposes. You can stop the monitor mode by writing an empty file with the name "stop.cryolo" in the input directory.',
    )

    optional_group.add_argument(
        "--otf",
        action="store_true",
        default=False,
        help="On the fly filtering. Filtered micrographs will not be written to disk. Might be slower",
    )

    optional_group.add_argument(
        "--cleanup",
        action="store_true",
        default=False,
        help="If true, it will delete the filtered images after training is done."
    )

    optional_group.add_argument(
        "--skip",
        action="store_true",
        default=False,
        help="If true, it will skip images that were already picked."
    )

    filament_group = parser.add_argument_group(
        "Filament options",
        "These options are only relevant if you want to use the filament mode.",
    )

    filament_group.add_argument(
        "--filament", action="store_true", help="Activate filament mode"
    )

    filament_group.add_argument(
        "-bd",
        "--box_distance",
        default=None,
        type=int,
        help="Distance in pixel between two boxes.",
    )

    filament_group.add_argument(
        "-mn",
        "--minimum_number_boxes",
        default=None,
        type=int,
        help="Minimum number of boxes per filament.",
    )

    filament_group.add_argument(
        "-sm",
        "--straightness_method",
        default="LINE_STRAIGHTNESS",
        help="Method to measure the straightness of a line. LINE_STRAIGHTNESS divides the length "
             "from start to end by the accumulated length between adjacent boxes. RMS calculates the "
             "root mean squared deviation of the line points to line given by start and the endpoint "
             "of the filament. Adjust the straightness_threshold accordingly!",
        choices=["NONE", "LINE_STRAIGHTNESS", "RMSD"],
    )

    filament_group.add_argument(
        "-st",
        "--straightness_threshold",
        default=0.95,
        type=float,
        help="Threshold value for the straightness method. The default value works good for LINE_STRAIGHTNESS. Lines with a LINE_STRAIGHTNESS lower than this threshold get splitted. For RMSD, lines with a RMSD higher than this threshold will be splitted. A good value for RMSD is 20 percent of your filament width.",
    )

    filament_group.add_argument(
        "-sr",
        "--search_range_factor",
        type=float,
        default=1.41,
        help="The search range for connecting boxes is the box size times this factor.",
    )

    filament_group.add_argument(
        "-ad",
        "--angle_delta",
        type=float,
        default=10,
        help="Angle delta in degree. This value is good more or less straight filaments. More curvy filaments might require values around 20.",
    )

    filament_group.add_argument(
        "--directional_method",
        default="PREDICTED",
        help="",
        choices=["PREDICTED", "CONVOLUTION"],
    )

    filament_group.add_argument(
        "-fw",
        "--filament_width",
        default=None,
        type=int,
        help="(OPTIONAL) Filament width (in pixel). You only need to provide this information when you use the"
             "directional_method \"CONVOLUTION\".",
    )

    filament_group.add_argument(
        "-mw",
        "--mask_width",
        default=100,
        type=int,
        help="(OPTIONAL) Mask width (in pixel). A Gaussian filter mask is used to estimate the direction of the filaments. This parameter defines how elongated the mask is. The default value typically don't has to be changed. It is only used when you use the directional_method \"CONVOLUTION\".",
    )

    filament_group.add_argument(
        "--nosplit",
        action="store_true",
        help="(DEPRECATED) The filament mode does not split to curved filaments. It is deprecated, you can set straightness_method to None instead.",
    )

    filament_group.add_argument(
        "--nomerging",
        action="store_true",
        help="The filament mode does not merge filaments",
    )

    tomo_group = parser.add_argument_group(
        "Tomography options",
        "These options are only relevant if you want to pick tomograms",
    )

    tomo_group.add_argument(
        "--tomogram", action="store_true", help="Activate tomography picking mode."
    )

    tomo_group.add_argument(
        "-tsr",
        "--tracing_search_range",
        type=int,
        default=-1,
        help="Search range in pixel. On default it will choose 25 percent of the box size.",
    )

    tomo_group.add_argument(
        "-tmem",
        "--tracing_memory",
        type=int,
        default=0,
        help="The maximum number of frames during which a particle can vanish, "
                               "then reappear nearby, and be considered the same particle",
    )

    tomo_group.add_argument(
        "-mn3d",
        "--minimum_number_boxes_3d",
        type=int,
        default=2,
        help=" (filaments only) Minimum number of boxes per filament.",
    )

    tomo_group.add_argument(
        "-tmin",
        "--tracing_min_length",
        type=int,
        default=5,
        help=" (particles only) The minimum number of boxes in one trace to be considered as valid particle."
    )
    tomo_group.add_argument(
        "-twin",
        "--tracing_window_size",
        type=int,
        default=-1,
        help="(filaments only) Window width when averaging filament positions. Default (-1) will use the box size as window width."
    )

    tomo_group.add_argument(
        "-tedge",
        "--tracing_min_edge_weight",
        type=float,
        default=0.4,
        help="(filaments only) Number between 0 and 1. The edge weight describes the amount of overlapping between two filaments from different slices that belong to the same filament."
    )

    tomo_group.add_argument(
        "-tmerge",
        "--tracing_merge_thresh",
        type = float,
        default = 0.8,
        help = "(filaments only) Number between 0 and 1. Filaments that are overlaping more than this threshold are merged completely."
    )


    depexp_group = parser.add_argument_group(
        "Deprecated/Experimental/Special ",
        "Contains either deprecated / experimental or very special options.",
    )
    depexp_group.add_argument(
        "-p",
        "--patch",
        type=int,
        help="(DEPRECATED) Number of patches. Use the config file in case you want to use patches.",
    )

    depexp_group.add_argument(
        "--write_empty",
        action="store_true",
        help="Write empty box files when not particle could be found.",
    )


def get_parser():
    parser = GooeyParser(
        description="Pick particles with crYOLO on any dataset",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    create_parser(parser)
    return parser


def _main_():

    # if sys.argv[1] == "gui":

    if len(sys.argv) >= 2:
        if not "--ignore-gooey" in sys.argv:
            sys.argv.append("--ignore-gooey")

    # r'^\d+ particles are found in .* \( (\d+) % \)$'
    kwargs = {"terminal_font_family": "monospace", "richtext_controls": True}
    Gooey(
        main,
        program_name="crYOLO Predict",
        image_dir=os.path.join(os.path.abspath(os.path.dirname(__file__)), "../icons"),
        progress_regex=r"^.* \( Progress:\s+(-?\d+) % \)$",
        disable_progress_bar_animation=True,
        tabbed_groups=True,
        **kwargs
    )()


def main(args=None):
    import logging
    logging.getLogger('tensorflow').setLevel(logging.ERROR)
    start_method = "fork"
    import sys
    if sys.platform == "win32":
        start_method = "spawn"
    try:
        os_start_method = os.environ["CRYOLO_MP_START"]

        if os_start_method in ["spawn", "fork"]:
            start_method = os_start_method
    except:
        pass
    try:
        multiprocessing.set_start_method(start_method)
    except RuntimeError:
        pass
    import cryolo.utils as util

    util.check_for_updates()

    if args is None:
        parser = get_parser()
        args = parser.parse_args()

    if isinstance(args.gpu, list):
        args.gpu = [re.sub("[^0-9]", "", num) for num in args.gpu if re.sub("[^0-9]", "", num).isnumeric()]
        if len(args.gpu) == 1:
            num_gpus = 1
            if args.gpu[0] != "-1":
                str_gpus = args.gpu[0].strip().split(" ")
                num_gpus = len(str_gpus)
                os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(str_gpus)
        else:
            str_gpus = [str(entry) for entry in args.gpu]
            num_gpus = len(str_gpus)
            os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(str_gpus)
    else:
        num_gpus = 1
        if args.gpu != -1 and len(args.gpu) > 0:
            str_gpus = str(args.gpu)
            os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(str_gpus)

    config_proto_args = {}
    if args.num_cpu != -1:
        config_proto_args["intra_op_parallelism_threads"] =args.num_cpu
        config_proto_args["inter_op_parallelism_threads"] = args.num_cpu
    if args.gpu_fraction < 1.0 and args.gpu_fraction > 0.0:
        import tensorflow as tf
        from keras.backend.tensorflow_backend import set_session

        config = tf.ConfigProto(**config_proto_args)
        config.gpu_options.allow_growth = True
        config.gpu_options.per_process_gpu_memory_fraction = args.gpu_fraction
        set_session(tf.Session(config=config))
    else:
        import tensorflow as tf
        from keras.backend.tensorflow_backend import set_session

        config = tf.ConfigProto(
            **config_proto_args)
        config.gpu_options.allow_growth = True


        set_session(tf.Session(config=config))

    config_path = args.conf
    weights_path = args.weights
    input_path = args.input
    for path_index, path in enumerate(input_path):
        input_path[path_index] = os.path.realpath(path)
    obj_threshold = args.threshold
    min_distance = args.distance
    mask_width = args.mask_width
    prediction_batch_size = args.prediction_batch_size
    no_merging = args.nomerging
    otf = args.otf
    do_merging = True
    global MONITOR
    MONITOR = args.monitor
    norm_margin = args.norm_margin
    search_range_factor = args.search_range_factor
    angle_delta = args.angle_delta
    straightness_method_str = args.straightness_method
    if straightness_method_str == "RMSD":
        straightness_method = filament_tracer._get_rms
    elif straightness_method_str == "LINE_STRAIGHTNESS":
        straightness_method = filament_tracer._get_straightness
    else:
        straightness_method = None

    straightness_threshold = args.straightness_threshold

    directional_method = args.directional_method

    predict_3D = args.tomogram
    tomo_search_range = args.tracing_search_range
    tomo_memory = args.tracing_memory
    tomo_min_length = args.tracing_min_length
    tomo_filament_window_size = args.tracing_window_size
    tomo_min_edge_weight = args.tracing_min_edge_weight
    tomo_merge_thresh = args.tracing_merge_thresh
    filament_mode = args.filament

    #if filament_mode and predict_3D:
    #    print("Filament mode and tomogram prediction are not compatible. Exit.")
    #    sys.exit()

    if no_merging:
        do_merging = False

    nosplit = False
    if args.nosplit is not None:
        nosplit = args.nosplit

    if straightness_method is None:
        nosplit = True

    outdir = None
    if args.output is not None:
        outdir = str(args.output)
    write_empty = args.write_empty
    min_size = args.minsize
    max_size = args.maxsize
    num_cpus = int(multiprocessing.cpu_count() / 2)
    if args.num_cpu != -1:
        num_cpus = args.num_cpu

    with open(config_path) as config_buffer:
        try:
            config = json.load(config_buffer)
        except json.JSONDecodeError:
            print(
                "Your configuration file seems to be corrupted. Please check if it is valid."
            )
        config["model"]["input_size"] = config_tools.get_adjusted_input_size(config)

    # Setup log dir
    if "other" in config and "log_path" in config["other"]:
        log_path = config["other"]["log_path"]
    else:
        log_path = "logs/"
    if not os.path.exists(log_path):
        os.makedirs(log_path)

    # Setup logging
    import sys
    timestr = time.strftime("%Y%m%d-%H%M%S")
    # Write command to disk

    utils.write_command(
        os.path.join(log_path, "cmdlogs/", "command_predict_" + timestr + ".txt"),
        "cryolo_predict.py " + " ".join(sys.argv[1:]),
    )


    if filament_mode:
        if args.directional_method == "CONVOLUTION" and args.filament_width is None:
            sys.exit("Please specify your filament width ( -fw / --filament_width)")
        else:
            filament_width = args.filament_width
        if args.box_distance is None:
            sys.exit("Please specify your box distance ( -bd / --box_distance)")
        else:
            box_distance = args.box_distance
        minimum_number_boxes = 1
        if args.minimum_number_boxes is not None:
            minimum_number_boxes = args.minimum_number_boxes
        minimum_number_boxes_3d = args.minimum_number_boxes_3d

    if args.patch is not None and args.patch > 0:
        num_patches = int(args.patch)
    else:
        num_patches = config_tools.get_number_patches(config)

    # Get overlap patches
    overlap_patches = 0
    if "overlap_patches" in config["model"]:
        overlap_patches = int(config["model"]["overlap_patches"])
    elif "anchors" in config["model"]:
        if not len(config["model"]["anchors"]) > 2:
            overlap_patches = config["model"]["anchors"][0]

    #  Get normalization option
    normalization_string = "STANDARD"
    if "norm" in config["model"]:
        normalization_string = config["model"]["norm"]

    write_direct = True
    if predict_3D or filament_mode:
        write_direct = False
    try:
        picking_results, about = do_prediction(
            config_path=config_path,
            config_pre=config,
            weights_path=weights_path,
            input_path=input_path,
            obj_threshold=obj_threshold,
            num_patches=num_patches,
            filament_mode=filament_mode,
            write_empty=write_empty,
            overlap=overlap_patches,
            num_images_batch_prediction=prediction_batch_size,
            num_gpus=num_gpus,
            num_cpus=num_cpus,
            otf=otf,
            normalization=normalization_string,
            normalization_margin=norm_margin,
            write_direct=write_direct,
            min_distance=min_distance,
            outdir=outdir,
            min_size=min_size,
            max_size=max_size,
            skip_picked=args.skip,
        )
    except tf.errors.ResourceExhaustedError:
        print("############################")
        print("Not enough GPU memory. Try to reduce prediction batch size (-pbs). ")
        print("############################")
        sys.exit(0)

    export_box_size = config_tools.get_box_size(config)[0]
    if write_direct == False:
        del_not_fully_immersed, del_min_distance, del_min_size = picking_postprocessing(
            picking_results,
            export_box_size,
            min_distance,
            min_size=min_size,
            max_size=max_size,
        )

        if min_distance > 0:
            print(
                del_min_distance,
                "particles were filtered because of the distance threshold",
                min_distance,
            )

        print(
            "Deleted",
            del_not_fully_immersed,
            "particles as they were not fully immersed in the micrograph",
        )

        print(
            "Deleted",
            del_min_size,
            "particles as they were out of specified size range",
        )

    ###############################
    #   Filament Post Processing
    ###############################

    # 1. Build sets of images (size = number of processors)
    # 2. Enhance
    # 3. Filamental post processing
    use_est_angles = False
    if about["filament_model"] and directional_method=="PREDICTED":
        use_est_angles = True
    elif about["filament_model"] and directional_method=="CONVOLUTION":
        use_est_angles = False
        print("Use the old convolutional method to estimate the local filament direction. ")
    elif filament_mode and about["filament_model"]==False and directional_method=="PREDICTED":
        use_est_angles = False
        print("##########")
        print("The directional method \"" + directional_method + "\" can't be used as your model is not a filament model."
                                                               "You need to retrain your picking model. Fall back to"
                                                               " old directional method \"CONVOLUTION\".")
        print("##########")

        if args.filament_width is None:
            sys.exit("Please specify your filament width ( -fw / --filament_width)")

    if filament_mode:
        picking_result_with_boxes = []
        picking_result_no_boxes = []
        picking_result_with_boxes_subsets = []
        picked_filaments = 0

        for picking_result_micrograph in picking_results:
            if picking_result_micrograph["boxes"]:

                picking_result_with_boxes.append(picking_result_micrograph)
            else:
                picking_result_micrograph["filaments"] = []
                picking_result_no_boxes.append(picking_result_micrograph)

        if picking_result_with_boxes:
            image_width, image_height = imagereader.read_width_height(
                picking_result_with_boxes[0]["img_path"]
            )
            rescale_factor = 1024.0 / max(image_width, image_height)
            rescale_factor_x = 1024.0 / image_width
            rescale_factor_y = 1024.0 / image_height
            print("Start filament tracing")
            if use_est_angles == False:
                mask_creator = maskstackcreator.MaskStackCreator(
                    filament_width=filament_width * rescale_factor,
                    mask_size=1024,
                    mask_width=mask_width,
                    angle_step=2,
                    bright_background=True,
                )

                print("Initialisation mask stack")
                mask_creator.init()
                # Devide picking result into chunks

            number_processors = num_cpus


            # Make subsets per tomogram/micrograph
            picking_result_with_boxes_subsets = [
                picking_result_with_boxes[i : i + number_processors]
                for i in range(0, len(picking_result_with_boxes), number_processors)
            ]
            process_counter = 1

            # Parallel tracing

            search_radius_scaled = export_box_size * rescale_factor * search_range_factor
            print("Trace filaments in batches")
            for picking_result_subset in tqdm(picking_result_with_boxes_subsets):
                image_subset = [
                    picking_result_subset[i]["img_path"]
                    for i in range(0, len(picking_result_subset))
                ]

                boxes_subset = [
                    picking_result_subset[i]["boxes"]
                    for i in range(0, len(picking_result_subset))
                ]

                if use_est_angles == False:
                    print("Enhance!!")
                    enhanced_images_list = line_enhancer.enhance_images(
                        image_subset, mask_creator, num_cpus
                    )

                # global filament_tracers
                global filament_tracers
                filament_tracers = []
                for image_index, boxset in enumerate(boxes_subset):
                    if use_est_angles == False:
                        enhanced_images = enhanced_images_list[image_index]
                        angle_images = [
                            enhanced_images[i]["max_angle"] for i in range(len(enhanced_images))
                        ]
                    # if it is tomogram boxset will be a list of list. one for each frame.
                    # in case of micrograph, it will just be a list of bounding boxes
                    is_micrograph = isinstance(boxset[0],BoundBox)
                    if is_micrograph:
                        # Micrograph
                        angle_image_flipped = None
                        if use_est_angles == False:
                            angle_image_flipped = np.flipud(angle_images[image_index])

                        tracer = filament_tracer.FilamentTracer(
                                boxes=boxset,
                                orientation_image=angle_image_flipped,
                                search_radius=search_radius_scaled,
                                angle_delta=angle_delta,
                                rescale_factor=rescale_factor,
                                rescale_factor_x=rescale_factor_x,
                                rescale_factor_y=rescale_factor_y,
                                do_merging=do_merging,
                                box_distance=box_distance,
                                nosplit=nosplit,
                                straightness_method=straightness_method,
                                straightness_threshold=straightness_threshold
                            )
                        filament_tracers.append((image_index,0,tracer))
                    else:
                        # Tomogram
                        for frame_index, boxes in enumerate(boxset):
                            angle_image_flipped = None
                            if use_est_angles == False:
                                angle_image_flipped = np.flipud(angle_images[frame_index])

                            tracer = filament_tracer.FilamentTracer(
                                    boxes=boxes,
                                    orientation_image=angle_image_flipped,
                                    search_radius=search_radius_scaled,
                                    angle_delta=angle_delta,
                                    rescale_factor=rescale_factor,
                                    rescale_factor_x=rescale_factor_x,
                                    rescale_factor_y=rescale_factor_y,
                                    do_merging=do_merging,
                                    box_distance=box_distance,
                                    nosplit=nosplit,
                                    straightness_method=straightness_method,
                                    straightness_threshold=straightness_threshold
                                )

                            filament_tracers.append(
                                (image_index,frame_index,tracer)
                            )
                #pool = multiprocessing.Pool(processes=num_cpus)
                #print("NUM TRACERS:", len(filament_tracers), "CPUS", num_cpus)
                with multiprocessing.Pool(processes=num_cpus) as pool:
                    subset_new_filaments = pool.map(
                                trace_subset_filements,
                                range(len(filament_tracers)),
                                #chunksize=1,
                            )


                print("Tracing done")
                for image_index, frame_index, filaments in subset_new_filaments:

                    if "filaments" not in picking_result_subset[image_index]:
                        picking_result_subset[image_index]["filaments"] = {}

                    # Min number of boxes filter:
                    filaments = filament_tracer.filter_filaments_by_num_boxes(
                        filaments, minimum_number_boxes
                    )
                    picked_filaments += len(filaments)

                    if len(filaments) >= 1:
                        if int(frame_index) in picking_result_subset[image_index]["filaments"]:
                            print("Strange. There are multiple picks per frame. Stop")
                            import sys
                            sys.exit(0)
                        picking_result_subset[image_index]["filaments"][int(frame_index)] = filaments




                print("Total number of filaments picked so far: ", picked_filaments)
                process_counter += 1

        if write_empty:
            picking_result_with_boxes_subsets.append(picking_result_no_boxes)

        print("Total number of filaments picked: ", picked_filaments)


    ###############################
    #   Write bounding boxes
    ###############################
    if len(picking_results) > 0:
        write_size_distribution_to_disk(picking_results, os.path.join(outdir, "DISTR"))

    if predict_3D and filament_mode == False:
        picking_results = predict_3D_boxes_from_2D_boxes(picking_results=picking_results,
                                                         obj_threshold=obj_threshold,
                                                         tomo_min_length=tomo_min_length,
                                                         tomo_search_range=tomo_search_range,
                                                         tomo_memory=tomo_memory)

    if filament_mode:
        if predict_3D:
            print("Do 3D filament tracing")
            used_boxsize = export_box_size
            if  tomo_search_range == -1:
                tomo_search_range = used_boxsize/2

            if tomo_memory == -1:
                tomo_memory = 0 #max(1, used_boxsize // 10)

            if tomo_filament_window_size == -1:
                tomo_filament_window_size = int(used_boxsize)*3

            picking_result_with_boxes_subsets = predict_3D_filaments_from_2d_filaments(
                picking_results=picking_result_with_boxes_subsets,
                tomo_search_range=tomo_search_range,
                tomo_memory=tomo_memory,
                window_size=tomo_filament_window_size,
                min_edge_weight=tomo_min_edge_weight,
                box_distance=box_distance,
                minimum_number_boxes=minimum_number_boxes_3d,
                tomo_merge_threshold=tomo_merge_thresh
            )
            num_traced = 0
            for r in picking_result_with_boxes_subsets:
                for r2 in r:
                    if "filaments_traced" not in r2:
                        num_traced = -1
                        break
                    num_traced = num_traced + len(r2["filaments_traced"])
                if num_traced == -1:
                    break
            if num_traced > -1:
                print("Volume tracing found ", num_traced, " Filaments")

        write_filaments_to_disk(outdir, picking_result_with_boxes_subsets, write_empty, is_3d=predict_3D)
    elif write_direct == False:
        prediction_result_to_disk(outdir, picking_results, write_empty=write_empty)

    ###
    # WRITE CSPARC Files - DAMN STUPID CRYOLO SOFTWARE DESIGN :-/
    ###
    if predict_3D == False:
        all_coordinates = []
        all_micrographs = []
        if filament_mode:
            for picking_result_subset in tqdm(picking_result_with_boxes_subsets, desc="Write cryoSPARC coordinates"):
                for result in picking_result_subset:
                    if result["filaments"] or write_empty:
                        for frame in result["filaments"]:
                            all_coordinates.extend(result["filaments"][frame])
                            all_micrographs.extend([os.path.basename(result['img_path'])]*len(result["filaments"][frame]))

        else:
            all_coordinates = []
            all_micrographs = []
            for box_to_write in tqdm(picking_results,desc="Write cryoSPARC coordinates"):
                all_coordinates.extend(box_to_write["boxes"][0])
                all_micrographs.extend([os.path.basename(box_to_write['img_path'])]*len(box_to_write["boxes"][0]))
        write_cryosparc_spa_coords(outdir,
                                   coordinates=all_coordinates,
                                   micrograph_name=all_micrographs)



    if args.cleanup:
        print("#####################################")
        print("Delete filtered images...")
        if os.path.exists(config["model"]["filter"][-1]):
            import shutil
            shutil.rmtree(config["model"]["filter"][-1])
        print("Done")
        print("#####################################")


def write_size_distribution_to_disk(picking_results, output_folder=""):

    if not os.path.exists(output_folder):
        os.makedirs(output_folder)
    estimated_size = []
    confidence = []

    for box_to_write in picking_results:
        boxes_set = box_to_write["boxes_unfiltered"]
        for boxes in boxes_set:
            for box in boxes:
                est_width = box.meta["boxsize_estimated"][0]
                est_height = box.meta["boxsize_estimated"][1]
                avg_size = (est_height + est_width) // 2
                estimated_size.append(avg_size)
                confidence.append(box.c)

    #################################
    # Size distribution histogram
    #################################
    if len(estimated_size)>0:
        est_size_mean = int(np.mean(estimated_size))
        est_size_sd = int(np.std(estimated_size))
        est_size_25q = int(np.percentile(estimated_size, 25))
        est_size_50q = int(np.percentile(estimated_size, 50))
        est_size_75q = int(np.percentile(estimated_size, 75))
        print("#####################################")
        print("")
        print("## Particle diameter distribution ##")
        print("MEAN:", est_size_mean, "px")
        print("SD:", est_size_sd, "px")
        print("25%-Quantile:", est_size_25q, "px")
        print("50%-Quantile:", est_size_50q, "px")
        print("75%-Quantile:", est_size_75q, "px")

        import matplotlib as mpl

        mpl.use("Agg")
        import matplotlib.pyplot as pl

        mpl.rcParams["figure.dpi"] = 200
        mpl.rcParams.update({"font.size": 7})
        width = max(10, int((np.max(estimated_size) - np.min(estimated_size)) / 10))
        pl.hist(estimated_size, bins=width)
        pl.title("Particle diameter distribution")
        pl.xlabel("Partilce diameter [px] (Bin size: " + str(width) + "px )")
        pl.ylabel("Count")

        timestr = time.strftime("%Y%m%d-%H%M%S")
        import csv

        output_path_distr_raw_txt = os.path.join(
            output_folder, "size_distribution_raw_" + timestr + ".txt"
        )
        np.savetxt(output_path_distr_raw_txt, estimated_size, fmt="%d")
        output_path_distr_txt = os.path.join(
            output_folder, "size_distribution_summary_" + timestr + ".txt"
        )
        with open(output_path_distr_txt, "w", newline="") as file:
            writer = csv.writer(file)
            writer.writerow(["Metric", "Value"])
            writer.writerow(["MEAN", est_size_mean])
            writer.writerow(["SD", est_size_sd])
            writer.writerow(["Q25", est_size_25q])
            writer.writerow(["Q50", est_size_50q])
            writer.writerow(["Q75", est_size_75q])
        print("Wrote particle size distribution summary to", output_path_distr_txt)
        output_path_distr_img = os.path.join(
            output_folder, "size_distribution_" + timestr + ".png"
        )
        print("Wrote plot of particle size distribution to", output_path_distr_img)
        pl.savefig(output_path_distr_img)
        pl.close()

        #################################
        # Confidence distribution histogram
        #################################
        output_path_conf_img = os.path.join(
            output_folder, "confidence_distribution_" + timestr + ".png"
        )
        output_path_conf_raw = os.path.join(
            output_folder, "confidence_distribution_raw_" + timestr + ".txt"
        )
        output_path_conf_sum = os.path.join(
            output_folder, "confidence_distribution_summary_" + timestr + ".txt"
        )

        # RAW DATA
        np.savetxt(output_path_conf_raw, confidence, fmt="%1.2f")

        # SUMMARY
        conf_mean = np.mean(confidence)
        conf_sd = np.std(confidence)
        conf_25q = np.percentile(confidence, 25)
        conf_50q = np.percentile(confidence, 50)
        conf_75q = np.percentile(confidence, 75)

        with open(output_path_conf_sum, "w", newline="") as file:
            writer = csv.writer(file)
            writer.writerow(["Metric", "Value"])
            writer.writerow(["MEAN", conf_mean])
            writer.writerow(["SD", conf_sd])
            writer.writerow(["Q25", conf_25q])
            writer.writerow(["Q50", conf_50q])
            writer.writerow(["Q75", conf_75q])

        # PLOT
        width = max(10, int((np.max(confidence) - np.min(confidence)) / 0.05))
        pl.hist(confidence, bins=width)
        pl.title("Confidence distribution")
        bin_size_str = "{0:.2f}".format(((np.max(confidence) - np.min(confidence)) / width))
        pl.xlabel("Confidence (Bin size: " + bin_size_str + ")")
        pl.ylabel("Count")
        pl.savefig(output_path_conf_img)
        print("")
        print("## Particle confidence distribution ##")
        print("Wrote confidence distribution summary", output_path_conf_sum)
        print("Wrote confidence distribution to", output_path_conf_img)
        print("")
        print("#####################################")
        pl.close()

def write_filaments_to_disk(outdir, picking_results, write_empty=False, is_3d = False):

    for picking_result_subset in picking_results:

        for result in picking_result_subset:
            if result["filaments"] or write_empty:

                if is_3d == False:
                    for frame in result["filaments"]:
                        pth = result["pth"]
                        eman_helix_segmented_path = pth
                        eman_start_end = pth
                        star_segmented = pth[:-3] + "star"
                        cbox_segmented = pth[:-3] + "cbox"
                        if outdir is not None:
                            filename = os.path.basename(pth)
                            eman_helix_segmented_path = os.path.join(
                                outdir, "EMAN_HELIX_SEGMENTED", filename
                            )
                            eman_start_end = os.path.join(
                                outdir, "EMAN_START_END", filename
                            )
                            filename = os.path.splitext(filename)[0] + ".star"
                            star_segmented = os.path.join(
                                outdir, "STAR_FILAMENT_SEGMENTED", filename
                            )

                            #star_cryopsarc = os.path.join(
                            #    outdir, "CRYOSPARC", "cryosparc.star"
                            #)

                            filename = os.path.splitext(filename)[0] + ".cbox"
                            cbox_segmented = os.path.join(
                                outdir, "CBOX_FILAMENT_SEGMENTED", filename
                            )

                            cbox_unfitlered = os.path.join(
                                outdir, "CBOX", filename
                            )


                        if len(result["filaments"]) > 1:
                            root_ext = os.path.splitext(eman_helix_segmented_path)
                            eman_helix_segmented_path = root_ext[0] + "_slice" + str(frame).zfill(4) + \
                                                        root_ext[1]

                            root_ext = os.path.splitext(eman_start_end)
                            eman_start_end = root_ext[0] + "_slice" + str(frame).zfill(4) + \
                                             root_ext[1]

                            root_ext = os.path.splitext(star_segmented)
                            star_segmented = root_ext[0] + "_slice" + str(frame).zfill(4) + \
                                             root_ext[1]

                        if not os.path.exists(os.path.dirname(eman_helix_segmented_path)):
                            os.makedirs(os.path.dirname(eman_helix_segmented_path))

                        if not os.path.exists(os.path.dirname(eman_start_end)):
                            os.makedirs(os.path.dirname(eman_start_end))

                        if not os.path.exists(os.path.dirname(star_segmented)):
                            os.makedirs(os.path.dirname(star_segmented))

                        if not os.path.exists(os.path.dirname(cbox_segmented)):
                            os.makedirs(os.path.dirname(cbox_segmented))

                        if not os.path.exists(os.path.dirname(cbox_unfitlered)):
                            os.makedirs(os.path.dirname(cbox_unfitlered))

                       # if not os.path.exists(os.path.dirname(star_cryopsarc)):
                        #    os.makedirs(os.path.dirname(star_cryopsarc))

                        CoordsIO.write_eman1_helicon(
                            filaments=result["filaments"][frame],
                            path=eman_helix_segmented_path,
                            image_filename=os.path.basename(result["img_path"]),
                        )

                        CoordsIO.write_eman1_filament_start_end(
                            filaments=result["filaments"][frame], path=eman_start_end
                        )

                        CoordsIO.write_star_filemant_file(
                            filaments=result["filaments"][frame], path=star_segmented
                        )

                        CoordsIO.write_cbox_file(
                            coordinates=result["filaments"][frame],
                            path=cbox_segmented
                        )

                        CoordsIO.write_cbox_file(
                            coordinates=result["boxes_unfiltered"][frame],
                            path=cbox_unfitlered
                        )

                        #CoordsIO.write_cryosparc_star_file(
                        #    path=star_cryopsarc,
                        #    coordinates=result["filaments"][frame],
                        #    micrograph_name=os.path.basename(result["img_path"])
                        #)
                else:

                    root_ext = os.path.splitext(result["pth"])
                    cbox_segmented_path = root_ext[0] + ".cbox"

                    if outdir is not None:
                        filename = os.path.basename(cbox_segmented_path)
                        cbox_segmented_path_untraced = os.path.join(
                            outdir, "CBOX_FILAMENTS_UNTRACED", filename
                        )

                        cbox_unfitlered = os.path.join(
                            outdir, "CBOX", filename
                        )

                        cbox_segmented_path_traced = os.path.join(
                            outdir, "CBOX_FILAMENTS_TRACED", filename
                        )

                        coords_path_traced = os.path.join(
                            outdir, "COORDS_TRACED", os.path.basename(root_ext[0])+".coords"
                        )

                        coords_fid_traced = os.path.join(
                            outdir, "COORDS_TRACED_FID", os.path.basename(root_ext[0])+"_fid.coords"
                        )

                    if not os.path.exists(os.path.dirname(cbox_segmented_path_untraced)):
                        os.makedirs(os.path.dirname(cbox_segmented_path_untraced))

                    if not os.path.exists(os.path.dirname(cbox_unfitlered)):
                        os.makedirs(os.path.dirname(cbox_unfitlered))

                    # Set z for coordinates
                    all_untraced_filaments = []
                    all_untraced_boxes = []
                    for frame in result["filaments"]:
                        for filament in result["filaments"][frame]:
                            for box in filament.boxes:
                                box.depth = 1
                                box.z = frame
                            all_untraced_filaments.append(filament)

                    for frame,boxset in enumerate(result["boxes"]):
                        for box in boxset:
                            box.depth = 1
                            box.z = frame
                            all_untraced_boxes.append(box)

                    CoordsIO.write_cbox_file(cbox_segmented_path_untraced, all_untraced_filaments)
                    CoordsIO.write_cbox_file(cbox_unfitlered, all_untraced_boxes)
                    if "filaments_traced" in result:

                        if not os.path.exists(os.path.dirname(cbox_segmented_path_traced)):
                            os.makedirs(os.path.dirname(cbox_segmented_path_traced))
                        CoordsIO.write_cbox_file(cbox_segmented_path_traced, result["filaments_traced"])
                        fil_boxes = []
                        fil_ids_array = [[],[],[],[]]

                        for fil_id, fil in enumerate(result["filaments_traced"]):
                            xs = [box.x + box.w/2 for box in fil.boxes]
                            ys = [box.y + box.h/2 for box in fil.boxes]
                            zs = [box.z for box in fil.boxes]
                            fil_boxes.extend(fil.boxes)

                            fil_ids_array[0].extend(xs)
                            fil_ids_array[1].extend(ys)
                            fil_ids_array[2].extend(zs)
                            fil_ids_array[3].extend([fil_id+1] * len(fil.boxes)) # FID needs be >1, otherweise relion will not recognize it.


                        if not os.path.exists(os.path.dirname(coords_path_traced)):
                            os.makedirs(os.path.dirname(coords_path_traced))
                        CoordsIO.write_coords_file(coords_path_traced, fil_boxes)

                        if not os.path.exists(os.path.dirname(coords_fid_traced)):
                            os.makedirs(os.path.dirname(coords_fid_traced))

                        np.savetxt(coords_fid_traced, np.array(fil_ids_array).T, fmt='%f')

def write_tomo_coords(outdir: str, box_to_write, write_empty: bool):
    original_path = box_to_write["pth"]
    eman_path = original_path
    box_set = box_to_write["boxes"]
    box_set_unfiltered = box_to_write["boxes_unfiltered"]

    if not write_empty:
        box_set = [boxes for boxes in box_set if len(boxes) > 0]
        box_set_unfiltered = [boxes for boxes in box_set_unfiltered if len(boxes) > 0]

    # Write eman 3d boxes
    for boxes_index, boxes in enumerate(box_set):

        if outdir is not None:
            filename = os.path.basename(eman_path)
            eman_path = os.path.join(outdir, "EMAN_3D", filename)

        # Create directory if it does not existes
        if not os.path.exists(os.path.dirname(eman_path)):
            os.makedirs(os.path.dirname(eman_path))

        CoordsIO.write_eman_boxfile3d(path=eman_path, boxes=boxes)

    # Write CBOX / coords 3D coordinates
    for boxes_index, boxes in enumerate(box_set):

        cbox_path = os.path.splitext(original_path)[0] + ".cbox"
        coords_path = os.path.splitext(original_path)[0] + ".coords"

        if outdir is not None:
            cbox_path = os.path.join(outdir, "CBOX_3D", os.path.basename(cbox_path))
            coords_path = os.path.join(outdir, "COORDS", os.path.basename(coords_path))

        # Create directory if it does not existes
        if not os.path.exists(os.path.dirname(cbox_path)):
            os.makedirs(os.path.dirname(cbox_path))
        if not os.path.exists(os.path.dirname(coords_path)):
            os.makedirs(os.path.dirname(coords_path))
        CoordsIO.write_cbox_file(path=cbox_path,
                                 coordinates=boxes)
        CoordsIO.write_coords_file(path=coords_path,
                                   boxes=boxes)

    # Write CBOX slices coordinates
    for boxes_index, boxes in enumerate(box_set_unfiltered):

        cbox_path = os.path.splitext(original_path)[0] + ".cbox"

        if outdir is not None:
            filename = os.path.basename(cbox_path)
            cbox_path = os.path.join(outdir, "CBOX_UNTRACED", filename)

        # Create directory if it does not existes
        if not os.path.exists(os.path.dirname(cbox_path)):
            os.makedirs(os.path.dirname(cbox_path))
        CoordsIO.write_cbox_file(path=cbox_path, coordinates=boxes)

def write_cryosparc_spa_coords(outdir: str, coordinates, micrograph_name):
    csparc_path = os.path.join(outdir, "CRYOSPARC", "cryosparc.star")
    os.makedirs(os.path.dirname(csparc_path), exist_ok=True)

    CoordsIO.write_cryosparc_star_file(path=csparc_path,
                                       coordinates=coordinates,
                                       micrograph_name=micrograph_name
                                       )
def write_spa_coords(outdir: str, box_to_write):
    original_path = box_to_write["pth"]
    eman_path = original_path
    box_set = box_to_write["boxes"]
    for boxes_index, boxes in enumerate(box_set):

        slices_index = ""
        if len(box_set) > 1:
            slices_index = "_slice" + str(boxes_index).zfill(4)

        star_path = os.path.splitext(original_path)[0] + slices_index + ".star"
        cbox_path = os.path.splitext(original_path)[0] + slices_index + ".cbox"

        ###
        # EMAN
        ###
        if outdir is not None:
            filename = os.path.basename(eman_path)
            eman_path = os.path.join(outdir, "EMAN", filename)
        # Create directory if it does not existes
        if not os.path.exists(os.path.dirname(eman_path)):
            os.makedirs(os.path.dirname(eman_path))
        CoordsIO.write_eman1_boxfile(path=eman_path, boxes=boxes)

        ###
        # STAR
        ###
        if outdir is not None:
            filename = os.path.basename(star_path)
            star_path = os.path.join(outdir, "STAR", filename)

        # Create directory if it does not existes
        if not os.path.exists(os.path.dirname(star_path)):
            os.makedirs(os.path.dirname(star_path))
        CoordsIO.write_star_file(path=star_path, boxes=boxes)

        ###
        # CBOX
        ###
        if outdir is not None:
            filename = os.path.basename(cbox_path)
            cbox_path = os.path.join(outdir, "CBOX", filename)

        # Create directory if it does not existes
        if not os.path.exists(os.path.dirname(cbox_path)):
            os.makedirs(os.path.dirname(cbox_path))
        CoordsIO.write_cbox_file(path=cbox_path, coordinates=box_to_write["boxes_unfiltered"][boxes_index])



def prediction_result_to_disk(outdir, picking_results, write_empty=False):
    for box_to_write in picking_results:
        if "is_3d" in box_to_write and box_to_write["is_3d"] == True:
            write_tomo_coords(outdir, box_to_write, write_empty)
        else:
            write_spa_coords(outdir, box_to_write)

def min_distance_filter(boxes, min_distance):
    min_distance_sq = min_distance * min_distance

    import itertools

    all_comb = list(itertools.combinations(boxes, 2))

    distsqs = list(itertools.starmap(utils.box_squared_distance, all_comb))
    low_distance_pairs = list(
        itertools.compress(all_comb, [distsq < min_distance_sq for distsq in distsqs])
    )
    for box_a, box_b in low_distance_pairs:
        box_to_delte = box_a
        if box_a.c > box_b.c:
            box_to_delte = box_b
        if box_to_delte in boxes:
            boxes.remove(box_to_delte)

    return boxes


def rescale(box, image_height, image_width, export_size=None):
    x_ll = box.x * image_width - box.w / 2.0 * image_height  # lower left
    y_ll = image_height - box.y * image_height - box.h / 2.0 * image_width # lower right

    if "angle" in box.meta:
        box.meta["angle"] = np.pi - box.meta["angle"]

    boxheight_in_pxl = box.h * image_width
    boxwidth_in_pxl = box.w * image_height
    if export_size is not None:
        delta_boxheight = export_size - boxheight_in_pxl
        delta_boxwidth = export_size - boxwidth_in_pxl
        x_ll = x_ll - delta_boxwidth / 2
        y_ll = y_ll - delta_boxheight / 2
        boxheight_in_pxl = export_size
        boxwidth_in_pxl = export_size
    box.x = x_ll
    box.y = y_ll

    box.w = boxwidth_in_pxl
    box.h = boxheight_in_pxl


    return box


def picking_postprocessing_single_res(
    boxes, img_path, export_size, min_distance, min_size=None, max_size=None
):
    # Rescaling
    image_width, image_height = imagereader.read_width_height(
        img_path
    )

    # Save estimated box size in meta data
    for box in boxes:
        box.meta["boxsize_estimated"] = (
            int(box.w * image_width),
            int(box.h * image_height),
        )
        #print(box.meta["boxsize_estimated"])

    # Resize box
    boxes = [
        rescale(box, image_height, image_width, export_size)
        for box in boxes]

    # Min distance filter
    del_min_distance = 0
    if min_distance > 0:
        del_min_distance = len(boxes)
        boxes = min_distance_filter(
            boxes, min_distance
        )
        del_min_distance = del_min_distance - len(boxes)

    # Filtering of particles which are not fully immersed in the micrograph
    del_not_fully_immersed = 0
    boxes_to_delete = get_not_fully_immersed_box_indices(
        boxes, image_height, image_width
    )
    for index in sorted(boxes_to_delete, reverse=True):
        del boxes[index]
    del_not_fully_immersed += len(boxes_to_delete)

    # Filtering according size
    out_of_size = 0

    out_of_size_box_indices = get_out_of_sizes_boxes(
        boxes, min_size=min_size, max_size=max_size
    )
    for index in sorted(out_of_size_box_indices, reverse=True):
        del boxes[index]
    out_of_size += len(out_of_size_box_indices)

    return boxes, del_not_fully_immersed, del_min_distance, out_of_size


def predict_3D_filaments_from_2d_filaments(
        picking_results,
        tomo_search_range,
        tomo_memory,
        window_size,
        min_edge_weight,
        box_distance,
        minimum_number_boxes=1,
        tomo_merge_threshold=0.8
):

    from cryolo import grouping3d
    for picking_result_subset in picking_results: # One for each image

        for result in picking_result_subset:
            filaments_traced_3d = grouping3d.do_tracing_filaments(
                filaments_dict=result["filaments"],
                search_range=tomo_search_range,
                memory=tomo_memory,
                window_size=window_size,
                min_edge_weight=min_edge_weight,
                box_distance=box_distance,
                resample_dist=box_distance,
                min_length_filament=minimum_number_boxes,
                merge_threshold=tomo_merge_threshold,
            )

            result["filaments_traced"] = filaments_traced_3d
    return picking_results


def predict_3D_boxes_from_2D_boxes(
        picking_results, obj_threshold, tomo_min_length, tomo_search_range=-1, tomo_memory=-1
):
    """
    :param picking_results:
    :param obj_threshold:
    :param tomo_min_length:
    :param tomo_search_range:
    :param tomo_memory:
    :return:
    """
    # create copy
    picking_results_copy = []
    from cryolo import grouping3d
    num_tomo_particles = 0
    num_particles = 0
    for result_ in picking_results:
        result = dict(result_)
        picking_results_copy.append(result)
        mean_box_size = np.mean(
            [np.mean([(box.w + box.h) / 2 for box in slice_boxes]) for slice_boxes in
             result["boxes"] if len(slice_boxes) > 0])
        if tomo_search_range == -1:
            tomo_search_range = max(1, mean_box_size // 4.0)
        if tomo_memory == -1:
            tomo_memory = max(1, mean_box_size // 10)
        num_particles += np.sum([len(slice_boxes) for slice_boxes in result["boxes"]])
        tracing_dict = {}
        assert len(result["boxes"]) == len(result[
                                               "slices_indices"]), "Slices indicies and number of box lists are not equal. stop."

        for k in range(len(result["boxes"])):
            tracing_dict[int(result["slices_indices"][k])] = result["boxes"][k]

        bounding_boxes_3d = grouping3d.do_tracing(
            boxes_dict=tracing_dict,
            search_range=tomo_search_range,
            memory=tomo_memory,
            min_length=tomo_min_length
        )

        bounding_boxes_3d = utils.non_maxima_suppress_fast_3d(bounding_boxes_3d,
                                                        nms_threshold=0.3,
                                                        obj_threshold=obj_threshold)

        for box in bounding_boxes_3d:
            box.depth = box.w  # its always a square box

        result["boxes"] = [bounding_boxes_3d]

        result["is_3d"] = True
        num_tomo_particles += len(bounding_boxes_3d)
        # Set the depth
        boxes_3D_per_slice = []
        for frame_index, boxes in enumerate(result["boxes_unfiltered"]):
            for box in boxes:
                box.z = frame_index
                box.depth = 1
                boxes_3D_per_slice.append(box)

        result["boxes_unfiltered"] = [boxes_3D_per_slice]
    print(num_tomo_particles, "Particles were found by tracing", num_particles,
          "through the volumes.")
    return picking_results_copy

def picking_postprocessing(
    picking_results, export_size, min_distance, min_size=None, max_size=None
):
    # Rescaling
    del_not_fully_immersed = 0
    del_min_distance = 0
    out_of_size = 0
    for picking_result_micrograph in picking_results:
        for boxes_index, boxes in enumerate(picking_result_micrograph["boxes"]):
            post_res = picking_postprocessing_single_res(
                boxes=boxes,
                img_path=picking_result_micrograph["img_path"],
                export_size=export_size,
                min_distance=min_distance,
                min_size=min_size,
                max_size=max_size
            )
            picking_result_micrograph["boxes"][boxes_index] = post_res[0]
            del_not_fully_immersed += post_res[1]
            del_min_distance += post_res[2]
            out_of_size += post_res[3]

            post_res2 = picking_postprocessing_single_res(
                boxes=picking_result_micrograph["boxes_unfiltered"][boxes_index],
                img_path=picking_result_micrograph["img_path"],
                export_size=export_size,
                min_distance=min_distance,
                min_size=None,
                max_size=None
            )
            picking_result_micrograph["boxes_unfiltered"][boxes_index] = post_res2[0]

    return del_not_fully_immersed, del_min_distance, out_of_size


def setup_prediction_monitor(input_path, img_paths):
    """
    Setups the monitor mode
    :param input_path: Folder to monitor
    :param img_paths: List that should be updated when a new file is popping up
    :return: None
    """

    # Setup the monitor mode
    def add_to_list(event):
        if os.path.basename(event.src_path).lower().startswith("stop.cryolo"):

            global MONITOR
            MONITOR = False
            os.remove(event.src_path)
        else:
            img_paths.append(event.src_path)

    patterns = [
        "*.tiff",
        "*.mrc",
        "*.tif",
        "*.mrcs",
        "*.png",
        "*.png",
        "*.jpeg",
        "*.cryolo",
    ]
    ignore_patterns = None  # ".*" #ignore invisible files
    ignore_directories = True
    case_sensitive = False
    image_dir_handler = PatternMatchingEventHandler(
        patterns, ignore_patterns, ignore_directories, case_sensitive
    )

    image_dir_handler.on_created = add_to_list

    go_recursively = False
    my_observer = Observer()
    print("Monitoring:", input_path[0])
    my_observer.schedule(
        image_dir_handler, os.path.realpath(input_path[0]), recursive=go_recursively
    )
    my_observer.start()
    return my_observer

def get_total_number_of_batches(img_pths, batchsize):
    num_frms = 0
    for pth in img_pths:
        num_frms += imagereader.get_num_frames(pth)
    return np.ceil(num_frms/batchsize)


def generate_batches(img_pths,
                     batchsize,
                     num_patches,
                     overlap,
                     otf = False,
                     config = None,
                     filter_options=None,
                     resize_to=None,
                     num_cpus=None,
                     ):
    """

    :param img_pths: Input image paths
    :param batchsize: Batch size
    :param num_patches: Number of patches
    :param overlap: Overlap of pathces
    :param otf: Indiciator (True/False) if on-the-fly processing is used
    :param config: Only needed in case if filtering (OTF or Monitor): crYOLO configuration dictory
    :param filter_options: Filtering options
    :param resize_to: In case of filtering, additional parameter for resizing
    :param num_cpus: Number of threads used for filtering
    :return:
        - batch_np - 3D numpy array with batch image data
        - tiles_coords - tile coordiantes incase of patching is used
        - slice_indicis - Tuple (a,b), where a is current slice indict and b is the total number of slices
        - img_size - Image size before patching
        - batch_img_paths - The paths of the images used in this batch.
    """
    skipped_images = []
    image_indices = []
    tiles_coords = []
    batch = []
    img_tiles = None
    slice_indices = []
    img_size = None
    batch_img_paths = []
    for current_index_image, img_pth in enumerate(img_pths):
        if os.path.basename(img_pth)[0] == ".":
            continue
        # Read image file!
        try:
            if (otf or MONITOR) and "filter" in config["model"]:
                image = do_filtering(
                    img_paths=[img_pth],
                    filter_options=filter_options,
                    resize_to=resize_to,
                    num_cpus=num_cpus,
                    return_np=True
                )[0]
            else:
                image = imagereader.image_read(img_pth)
        except ValueError:
            print("Image not valid: ", img_pth, "SKIPPED")
            skipped_images.append(img_pth)
            continue

        if len(image.shape) == 2:
            image = image[np.newaxis,...]

        if image is not None or image.shape[0] == 0 or image.shape[1] == 0:
            image_indices.append(current_index_image)

            for slice_index in range(image.shape[0]):
                slice = image[slice_index]
                if img_size == None:
                    img_size = slice.shape
                else:
                    if img_size != slice.shape:
                        print("It seems that crYOLO is picking images with different shape. That"
                              "should not happen. Image:", img_pth, slice_index)
                for patch_x in np.arange(0, num_patches):
                    for patch_y in np.arange(0, num_patches):
                        tile_coordinates = imagereader.get_tile_coordinates(
                            slice.shape[1],
                            slice.shape[0],
                            num_patches,
                            (patch_x, patch_y),
                            overlap=overlap,
                        )
                        tiles_coords.append(tile_coordinates)
                        img_tmp = slice[tile_coordinates[1], tile_coordinates[0]]

                        batch.append(img_tmp[np.newaxis,...])
                        slice_indices.append((slice_index+1,image.shape[0]))
                        batch_img_paths.append(img_pth)

                last_slice_on_last_image = slice_index==(image.shape[0]-1) and (len(img_pths)-1)==current_index_image
                if len(batch) == batchsize or last_slice_on_last_image:
                    batch_np = np.concatenate(batch)
                    yield batch_np, tiles_coords, slice_indices, img_size, batch_img_paths
                    batch = []
                    tiles_coords = []
                    slice_indices = []
                    batch_img_paths = []
        if MONITOR:
            while current_index_image == (len(img_pths)-1):
                import time
                time.sleep(1)
                if MONITOR == False:
                    break






def do_filtering(img_paths, filter_options, resize_to, num_cpus, return_np=False):
    if len(filter_options) > 2:
        if not return_np:
            print("Filter data using noise2noise model: ", filter_options[0])
            img_paths_filtered = utils.filter_images_noise2noise_dir(
                img_paths=img_paths,
                output_dir_filtered_imgs=filter_options[3],
                model_path=filter_options[0],
                padding=filter_options[1],
                batch_size=filter_options[2],
                resize_to=resize_to,
            )
            return img_paths_filtered
        else:
            filtered_images = []
            for img_pth in img_paths:
                image = utils.filter_image_noise2noise(
                    img_path=img_pth,
                    model_path=filter_options[0],
                    padding=filter_options[1],
                    batch_size=filter_options[2],
                )

                if resize_to is not None:
                    from PIL import Image

                    image = np.array(
                        Image.fromarray(image).resize(
                            resize_to, resample=Image.BILINEAR
                        )
                    )
                filtered_images.append(image)
            return filtered_images
    else:
        # Normal lowpass filter
        if not return_np:
            start_f = time.time()
            img_paths_filtered = utils.filter_images_lowpass(
                img_paths=img_paths,
                output_dir_filtered_imgs=filter_options[1],
                cutoff=filter_options[0],
                num_cpus=num_cpus,
                resize_to=resize_to,
            )
            end_f = time.time()
            print("Time needed for filtering:", end_f - start_f)
            return img_paths_filtered
        else:
            images = utils.filter_images_lowpass(
                img_paths=img_paths,
                output_dir_filtered_imgs=None,
                cutoff=filter_options[0],
                num_cpus=num_cpus,
                otf=True,
                resize_to=resize_to,
            )
            return images
    return img_paths_filtered


def get_prediction_image_list(input_path, ignore_pth=None):
    """
    Will find all images to pick in input_path. Will ignore those images for which already a
    boxfile exist in ignore_pth

    :param input_path: Path with images to predict
    :param ignore_pth:
    :return:
    """
    img_paths = []
    if isinstance(input_path, list):

        for path in input_path:
            isdir = os.path.isdir(path)
            if isdir:
                dir_files = os.listdir(path)

                dir_files = [
                    i
                    for i in dir_files
                    if not i.startswith(".")
                       and os.path.isfile(os.path.join(path, i))
                       and i.endswith(("tiff", "tif", "mrc", "mrcs", "png", "jpg", "jpeg", "rec"))
                ]

                img_paths.extend(
                    [os.path.join(path, image_file) for image_file in dir_files]
                )
            elif os.path.isfile(path):

                if path.endswith(".star"):
                    from pyStarDB import sp_pystardb as star
                    sfile = star.StarFile(path)
                    if 'micrographs' in sfile:
                        panda_dataframe = sfile['micrographs'] # With relion 3.1
                    else:
                        panda_dataframe = sfile[''] # the panda dataframe has noname (i.e.:data_)
                    if '_rlnMicrographName' in panda_dataframe.columns:
                        for pth in panda_dataframe._rlnMicrographName.unique():
                            if pth.endswith(
                                    ("tiff", "tif", "mrc", "png", "mrcs", "jpg", "jpeg", "rec")):
                                img_paths.append(pth)

                if not path.startswith(".") and path.endswith(
                        ("tiff", "tif", "mrc", "png", "mrcs", "jpg", "jpeg", "rec")
                ):
                    img_paths.append(path)
    else:
        isdir = os.path.isdir(input_path)
        if isdir:
            img_paths = os.listdir(input_path)

            img_paths = [
                i
                for i in img_paths
                if not i.startswith(".")
                   and os.path.isfile(os.path.join(input_path, i))
                   and i.endswith(("tiff", "tif", "mrc", "mrcs", "png", "jpg", "jpeg"))
            ]

    if ignore_pth is not None:
        import glob
        already_picked = glob.glob(os.path.join(ignore_pth, "*/*.box"))
        already_picked = [os.path.splitext(os.path.basename(box_path))[0] for box_path in
                          already_picked]
        img_paths = [img_path for img_path in img_paths if
                     os.path.splitext(os.path.basename(img_path))[0] not in already_picked]

    return img_paths

def do_prediction(
    config_path,
    weights_path,
    input_path,
    num_patches,
    obj_threshold=0.3,
    write_empty=False,
    config_pre=None,
    overlap=0,
    filament_mode=False,
    num_images_batch_prediction=3,
    num_gpus=1,
    num_cpus=-1,
    yolo=None,
    otf=False,
    normalization="STANDARD",
    normalization_margin=0,
    skip_picked=False,
    **kwargs
):
    """
    :param config_path: Path to the config file
    :param weights_path: Path do weights file (h5)
    :param input_path: Path to the folder containing the input images
    :param num_patches: Number of patches to use
    :param obj_threshold: Threshold for objects
    :param write_empty:
    :param config_pre:
    :param overlap: Overlap if patches are used
    :param filament_mode: If true, higher nms threshold is used to allow more overlapping.
    :param num_images_batch_prediction:
    :param num_gpus:
    :param yolo:
    :param otf: on the fly picking
    :param write_direct: on the fly writing of box files
    :return:
    1. List of dictionaris (on for each image) with the following keys:
        - pth - Potential path to the box file box_pth,
        - img_width - Original image width
        - img_height - Original image height
        - img_path - Oriignal image path
        - boxes - List of lists (one for each micrograph/slice) of Boxes filtered according confidence threshold
        - boxes_unfiltered - List of lists (one for each micrograph/slice) Boxes without applied confidence filtering
        - slices_indices - Slice indices for tomograms
    2. dictonary with information about training
    """
    for path in input_path:
        path_exists = os.path.exists(path)
        if not path_exists:
            sys.exit("Input path does not exist: " + path)

    if skip_picked:
        img_paths = get_prediction_image_list(input_path, kwargs["outdir"])
    else:
        img_paths = get_prediction_image_list(input_path)

    import time

    my_observer = None
    if MONITOR:
        my_observer = setup_prediction_monitor(input_path, img_paths)

    if not img_paths and MONITOR == False:
        sys.exit("No valid image in your specified input")

    if len(img_paths) == 0 and MONITOR == True:
        while len(img_paths) == 0:
            time.sleep(1)
        # When monitoring a directory, wait for the first valid image

    img_paths.sort()
    about = {}
    from cryolo.preprocessing import get_image_size_distr

    size_distr = get_image_size_distr(img_paths)
    if len(size_distr)>1:
        print("Prediction on mixed-size images is not supported in single prediction run.")
        print("The following input image sizes were detected:")
        for size in size_distr:
            print(size[0], "x", size[1], "( N:", size[2], ")")
        print("Stop crYOLO.")
        sys.exit(1)


    if config_pre is not None:
        config = config_pre.copy()
    else:
        with open(config_path) as config_buffer:
            config = json.load(config_buffer)
    cryolo_mode = utils.get_cryolo_mode(size_distr=size_distr, config_input_size=config["model"]["input_size"])
    if cryolo_mode == cryolo_mode.NON_SQUARE:
        ar = size_distr[0][0]/size_distr[0][1]
        if ar < 1:
            input_size = [int(config["model"]["input_size"] / ar), config["model"]["input_size"]]
        else:
            input_size = [config["model"]["input_size"], int(config["model"]["input_size"] * ar)]
        adjusted_size = config_tools.adjust_size(input_size)
        print("Your training input size",config["model"]["input_size"],"was adjusted to",adjusted_size)
        config["model"]["input_size"] = adjusted_size
        about["adjusted_input_size"] = adjusted_size

    # Read (first) image and check the image depth.
    first_image_path = img_paths[0]
    try:
        img_first = imagereader.image_read(first_image_path)
    except ValueError:
        sys.exit("Image " + first_image_path + " is not valid")
    if img_first is None:
        sys.exit("No valid image: " + first_image_path)

    if len(img_first.shape) == 2:
        depth = 1
    elif img_first.shape[2] == 1:
        depth = 1
    elif img_first.shape[2] > 3:
        depth = 1
    elif img_first.shape[2] == 3 and \
            np.all(img_first[:, :, 0] == img_first[:, :, 1]) and \
            np.all(
        img_first[:, :, 0] == img_first[:, :, 2]
    ):
        depth = 1
    else:
        depth = 3
    about["depth"] = depth
    #grid_w, grid_h = config_tools.get_gridcell_dimensions(config)

    #############################################
    # Read meta data about the model
    #############################################
    anchors = None
    yolo_kwargs = {}
    cryolo_version = None

    custom_params = utils.get_custom_paramters_from_model(weights_path)

    if "anchors" in custom_params:
        anchors = custom_params["anchors"]
    if "num_free_layers" in custom_params:
        yolo_kwargs["num_fine_tune_layers"] = custom_params["num_free_layers"]
    if "cryolo_version" in custom_params:
        cryolo_version = custom_params["cryolo_version"]
        about["model_cryolo_version"] = cryolo_version
    if "filament_model" in custom_params:
        filament_model = custom_params["filament_model"]
        about["filament_model"] = filament_model
    else:
        filament_model = False
        about["filament_model"] = False

    if cryolo_version is not None:
        print("Load crYOLO model which was trained with", cryolo_version)

    if anchors is None:
        #
        # TODO: Deprecated, anchors are now saved in the model and always defined.
        #
        anchors = utils.get_anchors(config, image_size=img_first.shape)
        print("Calculated Anchors using first image", anchors)
    else:
        print("Read Anchor from model", anchors)

    about["used_anchors"] = anchors

    if yolo is None:
        ###############################
        #   Make the model
        ###############################
        backend_weights = None
        if "backend_weights" in config["model"]:
            backend_weights = config["model"]["backend_weights"]
        from .frontend import YOLO

        yolo = YOLO(
            architecture=config["model"]["architecture"],
            input_size=config["model"]["input_size"],
            input_depth=depth,
            labels=["particle"],
            max_box_per_image=config["model"]["max_box_per_image"],
            anchors=anchors,
            backend_weights=backend_weights,
            pretrained_weights=weights_path,
            filament_model=filament_model,
            **yolo_kwargs
        )

        ###############################
        #   Load trained weights
        ###############################

        # USE MULTIGPU
        if num_gpus > 1:
            from keras.utils import multi_gpu_model

            parallel_model = multi_gpu_model(yolo.model, gpus=num_gpus)
            yolo.model = parallel_model
    else:
        yolo.anchors = anchors

    ##############################
    # Filter the data
    ##############################

    resize_to = None
    if num_patches == 1:
        # In case only one patch is used (should be default), the resizing can already
        # be done at the filtering step
        resize_to = config["model"]["input_size"]

    if MONITOR:
        # Monitor mode only works OTF
        otf = True

    if otf and not "filter" in config["model"]:
        print(
            "You specified the --otf option. However, filtering is not configured in your"
            "config line, therefore crYOLO will ignore --otf."
        )

    do_nn_filter = False
    filter_options = None
    if "filter" in config["model"]:
        filter_options = config["model"]["filter"]
        if otf == False:
            img_paths_filtered = do_filtering(
                img_paths=img_paths,
                filter_options=filter_options,
                resize_to=resize_to,
                num_cpus=num_cpus,
                         )
        else:
            img_paths_filtered = img_paths
    else:
        img_paths_filtered = img_paths
    ###############################
    #   Predict bounding boxes
    ###############################
    print("Reset progress bar: ( Progress: -1 % )")
    total_picked = 0
    boxes_to_write = []
    measured_times = []
    batchsize = num_patches * num_patches * num_images_batch_prediction
    skipped_images = []
    sum_del_not_fully_immersed = 0
    sum_del_min_distance = 0
    sum_del_out_of_size = 0
    boxes_per_image_dict = {}
    boxes_per_image_unfiltered_dict = {}

    total_num_batches = get_total_number_of_batches(img_paths_filtered,batchsize)

    batch_generator = generate_batches(img_paths_filtered, #pass a copy...
                         batchsize,
                         num_patches,
                         overlap,
                         otf,
                         config,
                         filter_options,
                         resize_to,
                         num_cpus,
                         )
    batch_num = 1
    if filament_mode:
        nms_thresh = 0.5
    else:
        nms_thresh = 0.3 # change back to 0.3 after testing!!!
    about["nms_thresh"] = nms_thresh

    for batch, tile_coord, slices_indicies, img_size, batch_img_paths in batch_generator:
        start = time.time()
        # Prediction

        boxes_per_image_nms_batch, boxes_per_image_unfiltered_batch = yolo.predict(
            batch,
            tile_coord,
            img_size,
            obj_threshold=obj_threshold,
            nms_threshold=nms_thresh,
            num_patches=num_patches,
            normalize_margin=normalization_margin,
            normalization=normalization,
        )

        # Now organize all boxes per image
        # In case of tomograms, there will be multiple box sets per image (on for each slice)
        particles_per_batch = 0

        images_completely_picked = []
        for batch_index, boxes in enumerate(boxes_per_image_nms_batch):
            particles_per_batch += len(boxes)
            key = os.path.basename(batch_img_paths[batch_index])
            if key in boxes_per_image_dict:
                boxes_per_image_dict[key]["boxes"].append(boxes)
                boxes_per_image_dict[key]["slice_indices"].append(slices_indicies[batch_index][0])
                boxes_per_image_unfiltered_dict[key]["boxes"].append(boxes_per_image_unfiltered_batch[batch_index])
            else:
                boxes_per_image_dict[key] = {
                    "boxes": [boxes],
                    "slice_indices": [slices_indicies[batch_index][0]]
                }
                boxes_per_image_unfiltered_dict[key] = {"boxes": [boxes_per_image_unfiltered_batch[batch_index]]}


            if slices_indicies[batch_index][0] == slices_indicies[batch_index][1]:
                images_completely_picked.append(key)


        # Generate datasets of boxes that contain all necessary meta data
        new_boxes_to_write = []
        if len(images_completely_picked) > 0:

            for preprocessed_img_pth in images_completely_picked:

                # Use original image data instead of filtered
                orig_img_path = utils.find_corresponding_path(img_paths,preprocessed_img_pth)

                # Use original image size instead of filtered image

                imgw, imgh = imagereader.read_width_height(
                    orig_img_path
                )

                box_pth = os.path.join(
                    os.path.dirname(orig_img_path),
                    "boxes",
                    os.path.splitext(os.path.basename(orig_img_path))[0]
                    + ".box",
                )

                boxes_filtered = boxes_per_image_dict[preprocessed_img_pth]["boxes"]
                boxes_unfiltered = boxes_per_image_unfiltered_dict[preprocessed_img_pth]["boxes"]

                boxes_on_one_image_to_write = {
                    "pth": box_pth,
                    "img_width": imgw,
                    "img_height": imgh,
                    "img_path": orig_img_path,
                    "boxes": boxes_filtered,
                    "boxes_unfiltered": boxes_unfiltered,
                    "slices_indices": boxes_per_image_dict[preprocessed_img_pth]["slice_indices"],
                }
                for boxes in boxes_filtered:
                    total_picked = total_picked + len(boxes)
                new_boxes_to_write.append(boxes_on_one_image_to_write)


        # Write to disk

        if "write_direct" in kwargs and len(images_completely_picked) > 0:
            if kwargs["write_direct"] == True:
                export_size = config_tools.get_box_size(config)[0]
                (
                    del_not_fully_immersed,
                    del_min_distance,
                    del_min_size,
                ) = picking_postprocessing(
                    picking_results=new_boxes_to_write,
                    export_size=export_size,
                    min_distance=kwargs["min_distance"],
                    min_size=kwargs["min_size"],
                    max_size=kwargs["max_size"],
                )

                sum_del_min_distance += del_min_distance
                sum_del_not_fully_immersed += del_not_fully_immersed

                sum_del_out_of_size += del_min_size

                prediction_result_to_disk(
                    kwargs["outdir"], new_boxes_to_write, write_empty=write_empty
                )
        boxes_to_write.extend(new_boxes_to_write)

        # Print progress
        prnt_progress = [
            particles_per_batch,
            "particles are found in image batch",
            batch_num,
            " / ",
            int(total_num_batches)
        ]

        if MONITOR == False:
            prnt_progress.append(" ( Progress: ")
            prnt_progress.append(
                int(float(batch_num) * 100 / total_num_batches)
            )

            prnt_progress.append("% )")
        end = time.time()
        measured_times.append(end - start)
        prnt_progress.append(f" Time: {measured_times[-1]:.2f}")
        print(*prnt_progress)
        batch_num += 1


    # Run post-processing
    if len(skipped_images) > 0:
        print(
            "The following images were skipped because of errors during reading them:"
        )
        for img in skipped_images:
            print(img)

    print("#####################################")
    print(
        total_picked,
        "particles in total has been found",
        "(",
        int(np.sum(measured_times)),
        "seconds)",
    )

    if "write_direct" in kwargs and kwargs["write_direct"] == True:
        if sum_del_min_distance > 0:
            print(
                sum_del_min_distance,
                "particles were filtered because of the distance threshold",
                kwargs["min_distance"],
            )

        print(
            "Deleted",
            sum_del_not_fully_immersed,
            "particles as they were not fully immersed in the micrograph",
        )
        print(
            "Deleted",
            sum_del_out_of_size,
            "particles as they were out of specified size range",
        )
    print("#####################################")
    return boxes_to_write, about


def get_out_of_sizes_boxes(boxes, min_size=None, max_size=None):
    if min_size is None and max_size is None:
        return []

    out_of_size_boxes = []
    for box_index, box in enumerate(boxes):
        box_width = box.meta["boxsize_estimated"][0]
        box_height = box.meta["boxsize_estimated"][1]
        size = (box_width + box_height) / 2
        out_of_size = False

        if min_size:
            if size <= min_size:
                out_of_size = True

        if max_size:
            if size >= max_size:
                out_of_size = True

        if out_of_size:
            out_of_size_boxes.append(box_index)

    return out_of_size_boxes


def get_not_fully_immersed_box_indices(boxes, image_height, image_width):
    boxes_to_delete = []
    for box_index, box in enumerate(boxes):
        box_width = box.w
        box_height = box.h
        if box_width == 0 and box_height == 0:
            box_width = box.meta["boxsize_estimated"][0]
            box_height = box.meta["boxsize_estimated"][1]
        if (
            (box.x + box_width) >= image_width
            or box.x < 0
            or box.y < 0
            or (box.y + box_height) >= image_height
        ):
            boxes_to_delete.append(box_index)


    return boxes_to_delete



def trace_subset_filements(i):
    image_index = filament_tracers[i][0]
    frame_index = filament_tracers[i][1]
    traceing_results =filament_tracers[i][2].trace_filaments()
    return (image_index,frame_index,traceing_results)


if __name__ == "__main__":
    _main_()



#
# COPYRIGHT
#
# All contributions by Thorsten Wagner:
# Copyright (c) 2017 - 2019, Thorsten Wagner.
# All rights reserved.
#
# ---------------------------------------------------------------------------
#         Do not reproduce or redistribute, in whole or in part.
#      Use of this code is permitted only under licence from Max Planck Society.
#            Contact us at thorsten.wagner@mpi-dortmund.mpg.de
# ---------------------------------------------------------------------------
#

from keras.callbacks import ModelCheckpoint


class MultiGPUModelCheckpoint(ModelCheckpoint):

    def __init__(self, *args, **kwargs):
        super(MultiGPUModelCheckpoint,self).__init__(*args,**kwargs)
        self.template_model = None
        self.fine_tune = None
        self.anchors = None
        self.num_free_layers = None
        self.version = None
        self.filament_model = False

    def set_template_model(self, template_model):
        self.template_model = template_model

    def set_fine_tune(self, fine_tune, num_free_layers):
        self.fine_tune = fine_tune
        self.num_free_layers = num_free_layers

    def set_anchors(self, anchors):
        self.anchors = anchors

    def set_filamet_model(self, is_filament_model):
        self.filament_model = is_filament_model

    def reset_best(self, best):
        self.best = best

    def set_version(self, version):
        self.version = version

    def on_epoch_end(self, epoch, logs=None):
        current = logs.get(self.monitor)
        if current is not None:
            if self.monitor_op(current, self.best):
                if self.verbose > 0:
                    print(
                        "\nEpoch %05d: %s improved from %0.5f to %0.5f,"
                        " saving model to %s"
                        % (epoch + 1, self.monitor, self.best, current, self.filepath)
                    )
                self.best = current
                if self.save_weights_only:
                    self.template_model.save_weights(self.filepath, overwrite=True)
                else:
                    self.template_model.save(self.filepath, overwrite=True)

                #############################################
                # Save meta data about the model
                #############################################
                import h5py

                with h5py.File(self.filepath, mode="r+") as f:
                    f["anchors"] = self.anchors
                    f["filament_model"] = [int(self.filament_model)]
                    if self.fine_tune:
                        f["num_free_layers"] = [self.num_free_layers]

                    if self.version is not None:
                        f["cryolo_version"] = [str(self.version).encode('utf8')]

            else:
                if self.verbose > 0:
                    print(
                        "\nEpoch %05d: %s did not improve from %0.5f"
                        % (epoch + 1, self.monitor, self.best)
                    )

    # def set_model(self, model):
    #    model_copy = clone_model(self.template_model)
    #    model_copy.set_weights(self.template_model.get_weights())
    #    self.model = model_copy
    # if isinstance(model.layers[-2], Model):
    #    self.model = model.layers[-2]
    # else:
    #    self.model = model



"""
Author: Thorsten Wagner (thorsten.wagner@mpi-dortmund.mpg.de)
"""
#
# COPYRIGHT
# All contributions by Thorsten Wagner:
# Copyright (c) 2017 - 2019, Thorsten Wagner.
# All rights reserved.
#
# ---------------------------------------------------------------------------
#         Do not reproduce or redistribute, in whole or in part.
#      Use of this code is permitted only under licence from Max Planck Society.
#            Contact us at thorsten.wagner@mpi-dortmund.mpg.de
# ---------------------------------------------------------------------------
#

import random
import numpy as np
from cryolo.augmentation.GaussBlurAugmentation import GaussBlurAugmentation
from cryolo.augmentation.AverageBlurAugmentation import AverageBlurAugmentation
from cryolo.augmentation.AdditiveGaussianNoiseAugmentation import AdditiveGaussianNoiseAugmentation
from cryolo.augmentation.AddConstantAugmentation import AddConstantAugmentation
from cryolo.augmentation.ContrastNormalizationAugmentation import ContrastNormalizationAugmentation
from cryolo.augmentation.DropoutAugmentation import DropoutAugmentation
from cryolo.augmentation.MultiplyAugmentation import MultiplyAugmentation

class Augmentation:
    """
    Class for doing data augmentation
    """

    def __init__(self, is_grey=False):
        self.is_grey = is_grey

    def image_augmentation(self, image):
        """
        Applies random selection of data augmentations
        :param image:  Input image
        :return: Augmented image
        """

        gauss_blur = GaussBlurAugmentation(sigma_range=[0, 3])
        avg_blur = AverageBlurAugmentation(kernel_size=[2, 7])
        gauss_noise = AdditiveGaussianNoiseAugmentation(max_sigma_range_factor=0.05)
        add_constant = AddConstantAugmentation(scale=0.05)
        constrast_normalization = ContrastNormalizationAugmentation(alpha_range=[0.5, 2.0])
        dropout = DropoutAugmentation(ratio=[0.01, 0.1])
        multiply = MultiplyAugmentation(mult_range=[0.5, 1.5])
        augmentations = [
            gauss_noise.transform_image,
            add_constant.transform_image,
            constrast_normalization.transform_image,
            multiply.transform_image,
            dropout.transform_image,
        ]

        num_augs = np.random.randint(0, np.minimum(6, len(augmentations)))
        if num_augs > 0:

            if np.random.rand() > 0.5:
                augmentations.append(gauss_blur.transform_image)
            else:
                augmentations.append(avg_blur.transform_image)

            selected_augs = random.sample(augmentations, num_augs)
            image = image.astype(np.float32, copy=False)
            for sel_aug in selected_augs:
                image = sel_aug(image)
            #   print "Mean after", sel_aug, " sum: ", np.mean(image)
            if self.is_grey:
                min_img = np.min(image)
                max_img = np.max(image)
                image = ((image - min_img) / (max_img - min_img)) * 255
                #    image = np.clip(image, 0, 255)
                image = image.astype(np.uint8, copy=False)

        return image



    def multiply(self, image, mult_range=(0.5, 1.5)):
        """
        Multipy the input image by a random float.
        :param image: Input image
        :param mult_range: Range for random multiplier
        :return: multiplied image
        """

        rand_multiplyer = mult_range[0] + np.random.rand() * (
            mult_range[1] - mult_range[0]
        )
        np.multiply(image, rand_multiplyer, out=image)
        return image



import unittest
import cryolo.CoordsIO
import os
from cryolo.utils import Filament, BoundBox


class CoordsIOTest(unittest.TestCase):

    def test_read_cbox_particle(self):
        path = os.path.join(os.path.dirname(__file__), "../resources/TcdA1-0019_frames_sum.cbox") #its a filament file
        part = cryolo.CoordsIO.read_cbox_boxfile(path)
        self.assertEqual(len(part), 129)

    def test_read_cbox_incomplete(self):
        path = os.path.join(os.path.dirname(__file__), "../resources/incomplete.cbox") #its a filament file
        fils = cryolo.CoordsIO.read_cbox_boxfile(path)
        self.assertEqual(len(fils), 12)

    def test_read_cbox_filament(self):
        path = os.path.join(os.path.dirname(__file__), "../resources/filament.cbox") #its a filament file
        fils = cryolo.CoordsIO.read_cbox_boxfile(path)
        self.assertEqual(len(fils), 78)

    def test_read_eman1_boxfile_lencorrect(self):
        path = os.path.join(os.path.dirname(__file__), "../resources/gammaADP-3200.box")
        boxes = cryolo.CoordsIO.read_eman1_boxfile(path)
        self.assertEqual(len(boxes), 21)

    def test_read_eman1_boxfile_last_box_correct(self):
        path = os.path.join(os.path.dirname(__file__), "../resources/gammaADP-3200.box")
        boxes = cryolo.CoordsIO.read_eman1_boxfile(path)

        self.assertEqual(boxes[len(boxes) - 1].x, 2177)
        self.assertEqual(boxes[len(boxes) - 1].y, 300)
        self.assertEqual(boxes[len(boxes) - 1].w, 256)
        self.assertEqual(boxes[len(boxes) - 1].h, 256)

    def test_write_eman1_boxfile(self):
        box_written = []
        for i in range(250, 750, 8):
            box_written.append(BoundBox(x=512, y=i, w=30, h=30))

        path = os.path.join(
            os.path.dirname(__file__), "../resources/test_write_eman1_boxfile.box"
        )

        cryolo.CoordsIO.write_eman1_boxfile(boxes=box_written, path=path)

        box_read = cryolo.CoordsIO.read_eman1_boxfile(path)

        for i, box in enumerate(box_written):
            self.assertEqual(box.x, box_read[i].x)
            self.assertEqual(box.y, box_read[i].y)
            self.assertEqual(box.w, box_read[i].w)
            self.assertEqual(box.h, box_read[i].h)

    def test_read_eman1_helicon_file3_onlyonefilament(self):
        path = os.path.join(
            os.path.dirname(__file__),
            "../resources/actin_cAla_1_corrfull_ptcl_coord_onlyone.txt",
        )
        filaments = cryolo.CoordsIO.read_eman1_helicon(path)
        print("NUM", len(filaments[0].boxes))
        self.assertEqual(1, len(filaments))

    def test_read_eman1_helicon_file3_onlyonefilament_numboxes(self):
        path = os.path.join(
            os.path.dirname(__file__),
            "../resources/actin_cAla_1_corrfull_ptcl_coord_onlyone.txt",
        )
        filaments = cryolo.CoordsIO.read_eman1_helicon(path)
        self.assertEqual(14, len(filaments[0].boxes))

    def test_read_eman1_helicon_file1(self):

        path = os.path.join(
            os.path.dirname(__file__),
            "../resources/actin_cAla_1_corrfull_ptcl_coord.txt",
        )
        filaments = cryolo.CoordsIO.read_eman1_helicon(path)
        self.assertEqual(4, len(filaments))

    def test_read_eman1_helicon_file2(self):

        path = os.path.join(
            os.path.dirname(__file__), "../resources/Actin-ADP-BeFx_0001.box"
        )
        filaments = cryolo.CoordsIO.read_eman1_helicon(path)
        self.assertEqual(23, len(filaments))

    def test_write_eman1_helicon_file1(self):
        boxes = []
        for i in range(250, 750, 8):
            boxes.append(BoundBox(x=512, y=i, w=30, h=30))

        filaments = [Filament(boxes)]

        path = os.path.join(
            os.path.dirname(__file__),
            "../resources/test_write_eman1_heliction_file1.box",
        )
        cryolo.CoordsIO.write_eman1_helicon(filaments, path, image_filename="test.mrc")

        filaments_read = cryolo.CoordsIO.read_eman1_helicon(path)

        box_written = filaments[0].boxes
        box_read = filaments_read[0].boxes
        self.assertEqual(len(box_written), len(box_read))
        for i, box in enumerate(box_written):
            print(box.x, box_read[i].x, box.y, box_read[i].y)

        for i, box in enumerate(box_written):
            print(box.x, box_read[i].x)
            self.assertEqual(box.x, box_read[i].x)
            self.assertEqual(box.y, box_read[i].y)
            self.assertEqual(box.w, box_read[i].w)
            self.assertEqual(box.h, box_read[i].h)

    def test_read_eman1_filament_start_end_len_file1(self):
        path = os.path.join(os.path.dirname(__file__), "../resources/Myo5ADP_0682.txt")
        filaments = cryolo.CoordsIO.read_eman1_filament_start_end(path)
        self.assertEqual(6, len(filaments))

    def test_write_eman1_filament_start_end_len_correct(self):
        boxes = []
        for i in range(250, 750, 8):
            boxes.append(BoundBox(x=512, y=i, w=30, h=30))

        filaments_written = [Filament(boxes)]

        path = os.path.join(
            os.path.dirname(__file__),
            "../resources/test_write_eman1_filament_start_end.box",
        )
        cryolo.CoordsIO.write_eman1_filament_start_end(
            filaments=filaments_written, path=path
        )

        filaments_read = cryolo.CoordsIO.read_eman1_filament_start_end(path)
        self.assertEqual(len(filaments_written), len(filaments_read))

    def test_write_eman1_filament_start_end_box_correct(self):
        box_written = []
        for i in range(250, 750, 8):
            box_written.append(BoundBox(x=512, y=i, w=30, h=30))

        filaments_written = [Filament(box_written)]

        path = os.path.join(
            os.path.dirname(__file__),
            "../resources/test_write_eman1_filament_start_end.box",
        )
        cryolo.CoordsIO.write_eman1_filament_start_end(
            filaments=filaments_written, path=path
        )

        filaments_read = cryolo.CoordsIO.read_eman1_filament_start_end(path)

        box_read = filaments_read[0].boxes

        self.assertEqual(box_read[0].x, box_written[0].x)
        self.assertEqual(box_read[0].y, box_written[0].y)
        self.assertEqual(box_read[0].w, box_written[0].w)
        self.assertEqual(box_read[0].h, box_written[0].h)

        self.assertEqual(
            box_read[len(box_read) - 1].x, box_written[len(box_written) - 1].x
        )
        self.assertEqual(
            box_read[len(box_read) - 1].y, box_written[len(box_written) - 1].y
        )
        self.assertEqual(
            box_read[len(box_read) - 1].w, box_written[len(box_written) - 1].w
        )
        self.assertEqual(
            box_read[len(box_read) - 1].h, box_written[len(box_written) - 1].h
        )

    def test_read_star_file_num_boxes(self):
        path = os.path.join(
            os.path.dirname(__file__), "../resources/TcdA1-0001_frames_sum.star"
        )
        boxes = cryolo.CoordsIO.read_star_file(path, 220)
        self.assertEqual(len(boxes), 128)

    def test_read_star_file_num_boxes_file2(self):
        path = os.path.join(os.path.dirname(__file__), "../resources/nompc.star")
        boxes = cryolo.CoordsIO.read_star_file(path, 220)
        self.assertEqual(len(boxes), 120)

    def test_read_star_file_first_box_correct(self):
        path = os.path.join(
            os.path.dirname(__file__), "../resources/TcdA1-0001_frames_sum.star"
        )
        boxes = cryolo.CoordsIO.read_star_file(path, 220)
        self.assertEqual(boxes[0].x + boxes[0].w / 2, 2302.0)
        self.assertEqual(boxes[0].y + boxes[0].h / 2, 4028.0)
        self.assertEqual(boxes[0].w, 220)
        self.assertEqual(boxes[0].h, 220)

    def test_read_star_file_last_box_correct(self):
        path = os.path.join(
            os.path.dirname(__file__), "../resources/TcdA1-0001_frames_sum.star"
        )
        boxes = cryolo.CoordsIO.read_star_file(path, 220)
        self.assertEqual(boxes[len(boxes) - 1].x + boxes[len(boxes) - 1].w / 2, 2602.5)
        self.assertEqual(boxes[len(boxes) - 1].y + boxes[len(boxes) - 1].w / 2, 90.5)
        self.assertEqual(boxes[len(boxes) - 1].w, 220)
        self.assertEqual(boxes[len(boxes) - 1].h, 220)

    def test_write_star_file(self):
        box_written = []
        for i in range(250, 750, 8):
            box_written.append(BoundBox(x=512, y=i, w=30, h=30))
        path = os.path.join(
            os.path.dirname(__file__), "../resources/test_write_star_file.star"
        )
        cryolo.CoordsIO.write_star_file(path, box_written)
        box_read = cryolo.CoordsIO.read_star_file(path, 30)

        for i, box in enumerate(box_written):
            self.assertEqual(box.x, box_read[i].x)
            self.assertEqual(box.y, box_read[i].y)
            self.assertEqual(box.w, box_read[i].w)
            self.assertEqual(box.h, box_read[i].h)


    def test_is_star_filament_file_singleparticle(self):
        path = os.path.join(os.path.dirname(__file__), "../resources/nompc.star")
        is_star_filament = cryolo.CoordsIO.is_star_filament_file(path)
        self.assertFalse(is_star_filament)

    def test_is_star_filament_file_singleparticle_file2(self):
        path = os.path.join(
            os.path.dirname(__file__), "../resources/normal_boxfile.box"
        )
        is_star_filament = cryolo.CoordsIO.is_star_filament_file(path)
        self.assertFalse(is_star_filament)

    def test_write_coords(self):
        path = os.path.join(os.path.dirname(__file__), "../resources/test.coords")
        a = cryolo.utils.BoundBox(x=0, y=0, z=0, w=5, h=5, depth=5)
        b = cryolo.utils.BoundBox(x=0, y=0, z=0, w=5, h=5, depth=5)
        boxes = [a, b]
        cryolo.CoordsIO.write_coords_file(path, boxes)
        # os.remove()



import unittest
from cryolo.evaluation.boxdataprovider import BoxDataProvider


class MyTestCase(unittest.TestCase):
    def test_group_method_lastfolder(self):
        group = BoxDataProvider.group_method_lastfolder(
            "/this/is/an/example/group/file.txt"
        )
        self.assertEqual("group", group)


if __name__ == "__main__":
    unittest.main()




from .boxdataprovider import BoxDataProvider, ImageBoxData
from cryolo.utils import BoundBox
from typing import Callable, List

class PathsDataProvider(BoxDataProvider):

    def __init__(self, paths : List[str],
                 reading_method : Callable[..., List[BoundBox]],
                 group_method : Callable[[str], str]):
        '''
        :param paths: List of filepaths.
        :param reading_method: Reading method for files in filepath.
        :param group_method: Extract the group out of the filename
        '''
        self.paths = paths
        self.reading_method = reading_method
        self.group_method = group_method

    def get_boxes(self) -> List[ImageBoxData]:
        result = []
        for file in self.paths:
            boxes = self.reading_method(file)
            group = self.group_method(file)
            arr = BoxDataProvider.boxes_to_nparray(boxes)
            dat = ImageBoxData(indentifer=file, data=arr, group=group)
            result.append(dat)

        return result



import numpy as np
from cryolo import utils
import tqdm
def convert_traces_to_bounding_boxes(traces):
    """
    Converts traces to a list of BoundingBox instances
    :param traces: Results of tracing
    :return: List of BoundingBox instances.
    """
    ccords = []
    for trace_id in np.unique(traces["particle"]):#range(num_particles):
        one_trace = traces[traces["particle"] == trace_id]

        xcoords = one_trace["x"].tolist()
        ycoords = one_trace["y"].tolist()
        zcoords = one_trace["frame"].tolist()
        widths = one_trace["widths"].tolist()
        heights = one_trace["heights"].tolist()
        confidence = one_trace["confidence"].tolist()
        est_box_widths = [meta["boxsize_estimated"][0] for meta in one_trace["meta"].tolist()]
        est_box_heights = [meta["boxsize_estimated"][1] for meta in one_trace["meta"].tolist()]

        x = np.mean(xcoords)
        y = np.mean(ycoords)
        z = np.mean(zcoords)
        w = np.mean(widths)
        h = np.mean(heights)
        c = np.mean(confidence)
        estimated_boxsize = (np.mean(est_box_widths),np.mean(est_box_heights))

        meta = {"boxsize_estimated": estimated_boxsize}
        meta["num_boxes"] = len(xcoords)
        bbox = utils.BoundBox(
            x=x,
            y=y,
            z=z,
            c=c,
            w=w,
            h=h,
            depth=w, # set it to same as the width. Should be a cube anyway.
            classes=[c]
        )
        bbox.meta = meta
        ccords.append(bbox)
    return ccords

def filaments_group_to_3D(filaments_df, window_size, box_distance):
    """
     For first filament: Subtract the first from
     last position and normalize the vector. This gives v0.
     v_avg = v0
     For each other filament
         Subtract(Last,First) and normalize -> v_k
         if v_k dot v_avg > 0:
           v_avg += v_k
           v_avg = norm(v_avg)
     Sort all filament positions along v_avg
     Running moving average along v_avg
     Resample


    :param filaments_df:
    :return:
    """

    def normalize_vec(vec):
        norm = np.linalg.norm(vec)
        if norm == 0:
            return vec
        return vec / norm

    def filament_direction_np(A):
        Acentered = A - A.mean(axis=0)
        Acov = np.zeros(shape=(2,2))
        #np.matmul(Acentered.T, Acentered, out=Acov)
        Acov = np.dot(Acentered.T, Acentered)
        evals, evecs = np.linalg.eigh(Acov)
        result = evecs[:, np.argmax(evals)]

        return result

    def filament_direction(fil):
        A = fil[["x","y"]]
        return filament_direction_np(A)

    nphelp = filaments_df["width"].to_numpy()
    same_box_size = (np.abs(nphelp-nphelp[0])<10**-6).all()

    assert same_box_size, print("The filaments have to have the same box size. Exit.")
    boxsize = nphelp[0]

    '''
    Calculate average direction v_avg in 2D of all filaments 
    '''
    fids = np.unique(filaments_df["fid"].tolist())
    pos = []
    for fid in fids:
        filament_trace = filaments_df[filaments_df["fid"] == fid]
        pos.append(filament_trace[["x","y"]])
    pos_help = np.concatenate(pos)
    v_avg = filament_direction_np( pos_help)

    '''
     Sort them according their x,y coordinates.
    '''
    xs = []
    ys = []
    zs = []
    conf=[]
    for fid in fids:
        filament_trace = filaments_df[filaments_df["fid"] == fid]
        xs.extend(filament_trace["x"].tolist())
        ys.extend(filament_trace["y"].tolist())
        zs.extend(filament_trace["frame"].tolist())
        conf.extend(filament_trace["confidence"].tolist())


    xs = np.array(xs)
    ys = np.array(ys)
    zs = np.array(zs)
    conf = np.array(conf)

    dotproducts = []
    for i in range(len(xs)):
        dp = np.dot(
            v_avg,
            np.array([xs[i], ys[i]])
        )
        dotproducts.append(
            dp
        )
    dotproducts = np.array(dotproducts)
    sorted_coords_indicis = np.argsort(dotproducts)
    xs = xs[sorted_coords_indicis]
    ys = ys[sorted_coords_indicis]
    zs = zs[sorted_coords_indicis]
    conf = conf[sorted_coords_indicis]
    dotproducts = dotproducts[sorted_coords_indicis]


    '''
    Moving average
    '''
    w = window_size
    boxes_3d = []

    for i in range(int(min(dotproducts)),int(max(dotproducts))+1):
        window_elements = np.logical_and(dotproducts > (i-w/2), dotproducts<i+w/2)

        if np.sum(window_elements) > 0:
            x3d = np.mean(xs[window_elements])
            y3d = np.mean(ys[window_elements])
            z3d = np.mean(zs[window_elements])
            conf3d = np.mean(conf[window_elements])

            box3d = utils.BoundBox(x=x3d,y=y3d,z=z3d, w=boxsize,h=boxsize,depth=boxsize,c=conf3d)
            box3d.meta["num_boxes"] = len(np.unique(zs[window_elements]))
            if len(boxes_3d)==0:
                boxes_3d.append(box3d)
            elif not np.allclose(
                    [boxes_3d[-1].x, boxes_3d[-1].y, boxes_3d[-1].z],
                    [x3d,y3d,z3d]
            ):
                boxes_3d.append(box3d)
            else:
                pass


    '''
    resample
    '''
    new_fil = utils.Filament(boxes=boxes_3d)
    resampled_filament = utils.resample_filaments([new_fil], box_distance)

    if len(resampled_filament) > 0:
        return resampled_filament[0]

    return None

def get_local_filament_direction(position, boxes, radius=2):

    from_i = max(0, position-radius)
    to_i = min(position+radius,len(boxes)-1)
    if from_i == to_i:
        return np.array([0,0])
    dx = boxes[from_i].x - boxes[to_i].x
    dy = boxes[from_i].y - boxes[to_i].y

    dvec = np.array([dx,dy])
    dvec = dvec / np.linalg.norm(dvec)
    return dvec
from typing import List, Tuple

def get_edge_histogram( edges: List[Tuple]) -> Tuple[np.array, List[Tuple]]:
    edgeset = set()
    edgeset.update(edges)
    bin_edges = list(edgeset)
    hist = np.zeros(len(bin_edges))
    for e in tqdm.tqdm(edges,desc="Calculate edge histogram"):
        i = bin_edges.index(e)
        hist[i] = hist[i] + 1

    return hist, bin_edges

def do_tracing_filaments(filaments_dict,
               search_range=5,
               memory=3,
               confidence_threshold=0,
               min_edge_weight=0,
               window_size=2,
               box_distance=1,
               resample_dist=-1,
               min_length_filament=0,
               iou_threshold=0.3,
               merge_threshold=0.8):
    '''

    :param filaments_dict: Filament dictionary. keys are the frame numbers
    :param search_range: Search range for tracing along z
    :param memory: Memory (allowed gap size during training)
    :param confidence_threshold: Confidence threshold
    :param min_edge_weight: Minimum edge weight. Edges shorts than this got removed
    :param window_size:
    :param box_distance:
    :param resample_dist:
    :param min_length_filament:
    :param iou_threshold:
    :param merge_threshold:
    :return:
    '''
    import pandas as pd
    import trackpy as tp
    import sys
    coords_and_frame = {
        "x": [],
        "y": [],
        "frame": [],
        "fid": [],
        "confidence":[],
        "width":[],
        "height":[],
        "meta":[],
        "fdirection":[]
    }
    fid = 0
    num_box = 0
    print("..Convert and resample")
    def printfilament(f):
        print("X", [b.x for b in f.boxes])
        print("Y", [b.y for b in f.boxes])
        print("Z", [b.y for b in f.boxes])

    for frame_number in filaments_dict:

        if resample_dist == -1:
            resampled_filaments = utils.resample_filaments(filaments_dict[frame_number],2*search_range+1)


        elif resample_dist > 0:

            resampled_filaments = utils.resample_filaments(filaments_dict[frame_number],
                                                           resample_dist)

        for filament_index, filament in enumerate(resampled_filaments):
            for box_i, box in enumerate(filament.boxes):
                if box.c > confidence_threshold:
                    coords_and_frame["x"].append(box.x)
                    coords_and_frame["y"].append(box.y)
                    coords_and_frame["frame"].append(frame_number)
                    coords_and_frame["fid"].append(fid)
                    coords_and_frame["confidence"].append(box.c)
                    coords_and_frame["meta"].append(box.meta)
                    coords_and_frame["width"].append(box.w)
                    coords_and_frame["height"].append(box.h)
                    coords_and_frame["fdirection"].append(get_local_filament_direction(position=box_i, boxes=filament.boxes, radius=2))
                    num_box = num_box + 1
            filaments_dict[frame_number][filament_index].meta["group3d_fid"] = fid
            fid = fid + 1
    if len(coords_and_frame["x"])==0:
        return []
    coords_frame_df = pd.DataFrame(coords_and_frame)


    tp.quiet(True)
    #import trackpy.linking as lnkng
    #lnkng.linking.Linker.MAX_SUB_NET_SIZE=50
    '''
    print("memory", memory)
    print("searchrange", search_range)
    print("windowsize", window_size)
    print("min_edge", min_edge_weight)
    print("conf", confidence_threshold)
    print("widht", coords_frame_df["width"][0])
    print("maxmin", np.max(coords_frame_df["x"]),np.min(coords_frame_df["x"]))
    '''
    #print("SUBETCONF:", lnkng.linking.Linker.MAX_SUB_NET_SIZE,lnkng.linking.Linker.MAX_SUB_NET_SIZE_ADAPTIVE, memory*search_range+1)
    print("..Trace", "SR", search_range, "MEM", memory)
    coords_frame_traced_df = tp.link_df(coords_frame_df,
                                        search_range=search_range,
                                        adaptive_stop = 2,
                                        adaptive_step = 0.8,
                                        memory=memory)

    #print("Linking done")
    print("..Build graph")
    # Next steps:
    # Build graph
    #np.unique(coords_frame_traced_df["particle"])
    import networkx as nx

    edges = []
    nodes = []
    all_r_values = []
    import tqdm

    for trace_id in tqdm.tqdm(np.unique(coords_frame_traced_df["particle"]), desc="Find nodes"):
        one_trace = coords_frame_traced_df[coords_frame_traced_df["particle"] == trace_id]

        '''
        filament_directions = one_trace["fdirection"].tolist()
        exceeded = False

        for i in range(len(filament_directions)-1):
            r = np.linalg.norm(np.cross(filament_directions[i], filament_directions[i+1]))
            all_r_values.append(r)
            if r > 0.15:
                exceeded = True
        if exceeded:
            pass
            #print("Exceeded!", r)
            #continue
        '''

        trace_fid_list = one_trace["fid"].tolist()
        for i in range(len(trace_fid_list)-1):
            edge = tuple(sorted((trace_fid_list[i],trace_fid_list[i+1])))
            edges.append(edge)
        nodes.extend(trace_fid_list)

    # Add nodes + Calculate edge weight + add weights
    nodes = set(nodes)

    G = nx.Graph()
    for node in nodes:
        G.add_node(node)

    thefids = coords_frame_traced_df["fid"]
    histfids, histfids_bins = np.histogram(thefids.values,bins=range(max(thefids)-min(thefids)+2))
    assert histfids_bins[0]==0
    hist_edges, hist_edges_bin = get_edge_histogram(edges)

    for edge in tqdm.tqdm(edges, desc="Build edges"):
        edge_weight = hist_edges[hist_edges_bin.index(edge)]#len([i for i,an_edge in enumerate(edges) if an_edge == edge])
        #print(edge_weight, hist_edges[hist_edges_bin.index(edge)])
        fil_length_a = histfids[edge[0]]#np.count_nonzero(thefids == edge[0])

        fil_length_b = histfids[edge[1]]#np.count_nonzero(thefids == edge[1])

        min_length = min(fil_length_a,fil_length_b)
        max_length = max(fil_length_a,fil_length_b)
        edge_weight = edge_weight/min_length * np.sqrt(min_length/max_length)
        has_minimum_weight = edge_weight >= min_edge_weight
        if has_minimum_weight:
            G.add_edge(edge[0],edge[1],weight=edge_weight)

    print("..Find connected components")
    # Find connected components
    connected_components = nx.connected_components(G)

    # Generate one 3D trace per connected component
    filaments_3d = []
    print("..Convert CCs to 3D Filaments")



    for sub in tqdm.tqdm(connected_components, desc="Convert to 3D"):
        filaments_of_cc = coords_frame_traced_df[ coords_frame_traced_df["fid"].isin(sub)]
        if len(filaments_of_cc.index) > 1:
            f3d = filaments_group_to_3D(filaments_of_cc, window_size=window_size, box_distance=box_distance)

            if f3d is not None:
                    filaments_3d.append(f3d)
                    f3d.meta["group3d_fids"] = sub

    '''
    Merge overlaping 3d filaments based on IOU
    '''
    print("..Merge strongly overlapping 3D Filaments", len(filaments_3d))
    filaments_3d = utils.merge_filaments_3d(filaments_3d,
                                            iou_threshold=iou_threshold,
                                            merge_threshold=merge_threshold,
                                            window_radius=window_size,
                                            box_distance=box_distance)
    print("..Merge partially ", len(filaments_3d))

    filaments_3d = utils.merge_filaments_3d(filaments_3d,
                                            iou_threshold=iou_threshold,
                                            merge_threshold=0,
                                            window_radius=window_size,
                                            box_distance=box_distance,
                                            partial=True)
    print("..Done", len(filaments_3d))
    print("Remove filaments shorter then:", min_length_filament)
    filaments_3d = [fil for fil in filaments_3d if len(fil.boxes)>=min_length_filament]

    print("..Done", len(filaments_3d))
    return filaments_3d


def do_tracing(boxes_dict,
               search_range=5,
               memory=3,
               confidence_threshold=0,
               min_length=5):
    """
    Will trace the boxes

    :param boxes_dict: Dictonary of lists of BoundingBox instances. Key the is the frame number.
    :param search_range: Maximum distance between position to get linked
    :param memory: Maximum gap in z direction.
    :param min_length: Minimum number of points (boxes in one  trace (default 5)
    :param confidence_threshold: Ignores boxes in boxes_list which have a confidence lower than this threshold
    :return: List of 3D bounding boxes
    """
    import pandas as pd
    import trackpy as tp
    coords_and_frame = {"x": [], "y": [], "frame": []}
    confidences = []
    meta = []
    widths = []
    heights = []
    for frame_number  in boxes_dict:
        for box in boxes_dict[frame_number]:
            if box.c > confidence_threshold:
                coords_and_frame["x"].append(box.x)
                coords_and_frame["y"].append(box.y)
                coords_and_frame["frame"].append(frame_number)
                confidences.append(box.c)
                meta.append(box.meta)
                widths.append(box.w)
                heights.append(box.h)
                assert len(box.classes) == 1, "for classes > 1 we need to complete the implementation"
    if len(coords_and_frame["x"])==0:
        return []
    coords_frame_df = pd.DataFrame(coords_and_frame)

    tp.quiet(True)
    coords_frame_traced_df = tp.link_df(coords_frame_df,
                                       search_range=search_range,
                                       memory=memory)
    coords_frame_traced_df["confidence"] = confidences
    coords_frame_traced_df["meta"] = meta
    coords_frame_traced_df["widths"] = widths
    coords_frame_traced_df["heights"] = heights
    coords_frame_traced_df= tp.filter_stubs(coords_frame_traced_df, min_length)
    boxes = convert_traces_to_bounding_boxes(coords_frame_traced_df)
    return boxes

def do_cluster(boxes_list,max_dist=2,max_z=3, confidence_threshold=0.3):

    num_boxes = np.sum([len([box for box in boxes if box.c>confidence_threshold]) for boxes in boxes_list])

    data = np.zeros(shape=(num_boxes,3))
    k = 0
    for boxes_index, boxes in enumerate(boxes_list):
        for box in boxes:
            if box.c > confidence_threshold:
                data[k,0]=box.x
                data[k,1]=box.y
                data[k,2]=boxes_index
                k = k + 1
    from sklearn import cluster
    epsilon = np.sqrt(max_dist * max_dist + max_z * max_z)

    clustering = cluster.AgglomerativeClustering(n_clusters=None, distance_threshold=80).fit(data) #cluster.DBSCAN(eps=epsilon, min_samples=3).fit(data)

    print("Num clusters found:", len(set(clustering.labels_)))

    return None#NotImplementedError("This method is not implemented yet")



import unittest
from cryolo.lowpass import zero_pad, next_power_of2
import numpy as np


class MyTestCase(unittest.TestCase):
    def test_zeropad_correct_size(self):

        img = np.zeros((5, 5)) + 1
        new_size = (10, 10)
        padded, _ = zero_pad(img, new_size)
        self.assertTupleEqual(padded.shape, new_size)

    def test_next_power_of2(self):

        num = next_power_of2(2 ** 2 + 1)
        self.assertEqual(num, 2 ** 3)


if __name__ == "__main__":
    unittest.main()



from gooey import Gooey, GooeyParser
import argparse
import os
import cryolo.config_tools as config_tools
import cryolo.__init__ as ini


def config_boxmanager_parser(parser_config):

    config_required_group = parser_config.add_argument_group(
        "General arguments", "Options to start the box manager.",
    )

    config_required_group.add_argument(
        "-i", "--image_dir", help="Path to image directory.", widget="DirChooser"
    )

    config_required_group.add_argument(
        "-b", "--box_dir", help="Path to box directory.", widget="DirChooser"
    )

    config_required_group.add_argument(
        "--wildcard", help="Wildcard for selecting specific images (e.g *_new_*.mrc)"
    )


def config_configuration_parser(parser_config):
    config_required_group = parser_config.add_argument_group(
        "General arguments",
        "These arguments most often have to be changed when creating a config file for crYOLO",
    )

    config_required_group.add_argument(
        "config_out_path",
        default="config_cryolo.json",
        help="Name of configuration file to create.",
        widget="FileSaver",
        gooey_options={
            "validator": {
                "test": 'user_input.endswith("json")',
                "message": "File has to end with .json!",
            },
            "default_file": "config_cryolo.json",
        },
    )
    config_required_group.add_argument(
        "boxsize",
        type=int,
        help="You should specify the same box size here as you used in your training data.  If you train on several datasets with varying box sizes, use something like the average of the used box sizes. In case of the general model fill in your target box size.",
    )

    config_required_group.add_argument(
        "--train_image_folder",
        help="Path to the image folder containing the images to train on. The field can only be left empty when you use a general model.",
        default="",
        widget="DirChooser",
    )
    config_required_group.add_argument(
        "--train_annot_folder",
        help="Path to folder containing the your annotation files like box or star files. The field can only be left empty when you use a general model.",
        default="",
        widget="DirChooser",
    )

    config_required_group.add_argument(
        "--saved_weights_name",
        default="cryolo_model.h5",
        help="Path for saving final weights.",
        widget="FileSaver",
        gooey_options={
            "validator": {
                "test": 'user_input.endswith("h5")',
                "message": "File has to end with .h5!",
            },
            "default_file": "cryolo_model.h5",
        },
    )

    config_model_group = parser_config.add_argument_group(
        "Model options",
        "Options to configure your model. The default options are well tested and typically don't have to be changed.",
    )

    config_model_group.add_argument(
        "-a",
        "--architecture",
        default="PhosaurusNet",
        choices=["PhosaurusNet", "YOLO", "crYOLO"],
        help="Backend network architecture.",
    )
    config_model_group.add_argument(
        "--input_size",
        default=1024,
        nargs="+",
        help="The shorter image dimension is downsized to this size and the long size according the aspect ratio. This is not the size of your micrographs and only rarely needs to be changed. You can also specify height and width of the input image size by seperating them by a whitespace.",
        gooey_options={
            "validator": {
                "test": "float(user_input.split()[0]) % 32 < 1 and float(user_input.split()[1]) % 32 < 1 if len(user_input.split()) == 2 else float(user_input.split()[0]) % 32 < 1",
                "message": "Input size has to be a multiple of 32!",
            }
        },
    )
    config_model_group.add_argument(
        "-nm",
        "--norm",
        default="STANDARD",
        help="Normalization that is applied to the images. Standard will subtract the image mean and devide by the standard deviation."
             " Experimental: Gaussian Mixture Models (GMM) fit a 2 component GMM to you image data and normalizes according the brighter component. "
             "This ensures that it always normalize with respect to ice but slows down the training.",
        choices=["STANDARD", "GMM"],
    )

    config_model_group.add_argument(
        "--num_patches",
        type=int,
        default=1,
        help='(DEPRECATED) If specified the patch mode will be used. A value of "2" means, that 2×2 patches will be used. A value of "1" means, that no patches will be used (recommended).',
    )
    config_model_group.add_argument(
        "--overlap_patches",
        type=int,
        default=200,
        help="(DEPRECATED) Only needed when using patch mode. Specifies how much the patches overlap. In our lab, we always keep the default value.",
    )

    config_denoising_group = parser_config.add_argument_group(
        "Denoising options",
        "By default crYOLO denoises your data before training/picking. It will use a low-pass filter, but you can also use neural network denoising (JANNI). Default options are good and typically don't have to be changed.",
    )

    config_denoising_group.add_argument(
        "--filtered_output",
        default="filtered_tmp/",
        help="Output folder for filtered images",
        widget="DirChooser",
    )

    config_denoising_group.add_argument(
        "-f",
        "--filter",
        default="LOWPASS",
        help="Noise filter applied before training/picking. You can choose between a normal low-pass filter"
        " and neural network denoising (JANNI).",
        choices=["NONE", "LOWPASS", "JANNI"],
    )

    config_denoising_group.add_argument(
        "--low_pass_cutoff",
        type=float,
        default=0.1,
        help="Low pass filter cutoff frequency",
        gooey_options={
            "validator": {
                "test": "0.0 <= float(user_input) <= 0.5",
                "message": "Must be between 0 and 0.5",
            }
        },
    )
    config_denoising_group.add_argument(
        "--janni_model",
        help="Path to JANNI model",
        widget="FileChooser",
        gooey_options={"wildcard": "*.h5"},
        default=None,
    )

    config_denoising_group.add_argument(
        "--janni_overlap",
        type=int,
        default=24,
        help="Overlap of patches in pixels (only needed when using JANNI)",
    )

    config_denoising_group.add_argument(
        "--janni_batches",
        type=int,
        default=3,
        help="Number of batches (only needed when using JANNI)",
    )

    config_training_group = parser_config.add_argument_group(
        "Training options",
        "These options influences your training. Default values are robust and typically don't have to be changed.",
    )

    config_training_group.add_argument(
        "--pretrained_weights",
        default="",
        help="Path to h5 file that is used for initialization in case you want to fine tune a previous model.",
        widget="FileChooser",
        gooey_options={
            "validator": {
                "test": 'user_input.endswith("h5")',
                "message": "File has to end with .h5!",
            },
            "wildcard": "*.h5",
        },
    )

    config_training_group.add_argument(
        "--train_times",
        default=10,
        type=int,
        help="How often each image is presented to the network during one epoch. The default should be kept until you have many training images.",
    )

    config_training_group.add_argument(
        "--batch_size",
        type=int,
        default=4,
        help="The number of images crYOLO process in parallel during training.",
    )
    config_training_group.add_argument(
        "--learning_rate",
        type=float,
        default=10 ** -4,
        help="Defines the step size during training. Default should be kept.",
    )
    config_training_group.add_argument(
        "--nb_epoch",
        type=int,
        default=200,
        help="Maximum number of epochs the network will train.",
    )
    config_training_group.add_argument(
        "--object_scale",
        type=float,
        default=5.0,
        help="Penalty scaling factor for missing picking particles.",
    )
    config_training_group.add_argument(
        "--no_object_scale",
        type=float,
        default=1.0,
        help="Penalty scaling factor for picking background.",
    )
    config_training_group.add_argument(
        "--coord_scale",
        type=float,
        default=1.0,
        help="Penalty scaling factor for errors in estimating the correct position.",
    )
    config_training_group.add_argument(
        "--class_scale",
        type=float,
        default=1.0,
        help='Irrelevant, as crYOLO only has the class "particle".',
    )

    config_training_group.add_argument(
        "--debug",
        action="store_true",
        default=True,
        help="If true, the network will provide several statistics during training.",
    )

    config_validation_group = parser_config.add_argument_group(
        "Validation configuration",
        "Validation configuration. If not specified, crYOLO will simply select 20% of the training "
        "data for validation. However it is possible to specify to use specific images "
        "for validation. If you do so, it is important to remove them from your training directories!",
    )
    config_validation_group.add_argument(
        "--valid_image_folder",
        default="",
        help="Path to folder containing the image files.",
        widget="DirChooser",
    )
    config_validation_group.add_argument(
        "--valid_annot_folder",
        default="",
        help="Path to folder containing the validation box files",
        widget="DirChooser",
    )

    config_other_group = parser_config.add_argument_group(
        "Other option", "Other options that does not fit anywhere else",
    )

    config_other_group.add_argument(
        "--log_path", default="logs/", help="Path for log saving", widget="DirChooser"
    )


def create_parser():
    parent_parser = GooeyParser(
        description="The crYOLO particle picking procedure!",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    subparsers = parent_parser.add_subparsers(help="sub-command help")

    # Config generator
    parser_config = subparsers.add_parser(
        "config",
        help="Command to create the configuration file",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    config_configuration_parser(parser_config)

    # Training parser
    parser_train = subparsers.add_parser(
        "train",
        help="Command to train crYOLO",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    from cryolo.train import create_parser as create_train_parser

    create_train_parser(parser_train)

    # Picking parser
    parser_pick = subparsers.add_parser(
        "predict",
        help="Command to pick particles using a trained model.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    # import cryolo.predict as predict
    from cryolo.predict import create_parser as create_predict_parser

    create_predict_parser(parser_pick)

    # Evaluation parser
    parser_eval = subparsers.add_parser(
        "evaluation",
        help="Command to evaluate the picking.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    from cryolo.eval import create_parser as create_eval_parser

    # import cryolo.eval as evaluation

    create_eval_parser(parser_eval)

    # crYOLO boxmanager
    parser_config = subparsers.add_parser(
        "boxmanager",
        help="Display boxes",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    config_boxmanager_parser(parser_config)

    return parent_parser


PARSER = None


def main():
    import sys

    args = PARSER.parse_args()

    print("#####################################################")
    print(
        "Important debugging information.\nIn case of any problems, please provide this information."
    )
    print("#####################################################")
    for cmd_arg in sys.argv:
        if cmd_arg.startswith("-"):
            print("")
        print(cmd_arg, end=" ")
    print("")
    print("#####################################################")

    if "config" in sys.argv[1]:
        import cryolo.utils as util

        util.check_for_updates()

        filter = None
        if args.filter == "LOWPASS":
            filter = [args.low_pass_cutoff, args.filtered_output]
        elif args.filter == "JANNI":
            if args.janni_model is None:
                print("Please specify the JANNI model file.")
                sys.exit(1)
            filter = [
                args.janni_model,
                args.janni_overlap,
                args.janni_batches,
                args.filtered_output,
            ]
        if type(args.input_size) == list:
            if len(args.input_size) == 2:
                args.input_size = [int(args.input_size[0]),int(args.input_size[1])]
            elif len(args.input_size) == 1:
                args.input_size = int(args.input_size[0])
        config_tools.generate_config_file(
            config_out_path=args.config_out_path,
            architecture=args.architecture,
            input_size=args.input_size,
            anchors=[args.boxsize, args.boxsize],
            max_box_per_image=700,
            num_patches=args.num_patches,
            overlap_patches=args.overlap_patches,
            filter=filter,
            train_image_folder=args.train_image_folder,
            train_annot_folder=args.train_annot_folder,
            train_times=args.train_times,
            pretrained_weights=args.pretrained_weights,
            batch_size=args.batch_size,
            learning_rate=args.learning_rate,
            nb_epoch=args.nb_epoch,
            object_scale=args.object_scale,
            no_object_scale=args.no_object_scale,
            coord_scale=args.coord_scale,
            class_scale=args.class_scale,
            log_path=args.log_path,
            saved_weights_name=args.saved_weights_name,
            debug=args.debug,
            valid_image_folder=args.valid_image_folder,
            valid_annot_folder=args.valid_annot_folder,
            valid_times=1,
            normalization=args.norm
        )
    elif "train" in sys.argv[1]:
        import cryolo.train as train

        train.main(args)
    elif "predict" in sys.argv[1]:
        import cryolo.predict as predict

        predict.main(args)
    elif "evaluation" in sys.argv[1]:
        import cryolo.eval as evaluation

        evaluation.main(args)
    elif "boxmanager" in sys.argv[1]:

        import os
        cmd = "napari_boxmanager"
        if args.image_dir:
            if args.wildcard:
                args.image_dir = os.path.join(args.image_dir,args.wildcard)
            cmd = cmd + f" '{args.image_dir}'"
        if args.box_dir:
            cmd = cmd + f" '{args.box_dir}'"
        os.system(cmd)

        #from cryoloBM import boxmanager

        #boxmanager.start_boxmanager(args.image_dir, args.box_dir, args.wildcard, is_tomo=args.is_tomo_dir)


def _main_():
    global PARSER
    import sys

    PARSER = create_parser()
    # try:
    #     multiprocessing.set_start_method("spawn")
    # except RuntimeError:
    #     print("Ignore set start method")

    kwargs = {"terminal_font_family": "monospace", "richtext_controls": True}

    if len(sys.argv) >= 2:
        if not "--ignore-gooey" in sys.argv:
            sys.argv.append("--ignore-gooey")

    Gooey(
        main,
        program_name="crYOLO " + ini.__version__,
        image_dir=os.path.join(os.path.abspath(os.path.dirname(__file__)), "../icons"),
        progress_regex=r"^.* \( Progress:\s+(-?\d+) % \)$",
        disable_progress_bar_animation=True,
        tabbed_groups=True,
        default_size=(1024, 630),
        **kwargs
    )()


if __name__ == "__main__":
    # try:
    #    multiprocessing.set_start_method("spawn")
    # except RuntimeError:
    #    print("Ignore set start method")
    _main_()



import unittest
from cryolo.evaluation.recallprecisionmetric3d import RecallPrecisionMetric3D
import numpy as np
import copy


class MyTestCase(unittest.TestCase):
    def test_precision_available(self):

        predicted = np.random.random(size=(10, 6))

        gt = np.random.random(size=(10, 6))
        eval = RecallPrecisionMetric3D()
        results = eval.eval(predicted, gt)

        self.assertTrue("precision" in results)

    def test_recall_available(self):

        predicted = np.random.random(size=(10, 6))

        gt = np.random.random(size=(10, 6))
        eval = RecallPrecisionMetric3D()
        results = eval.eval(predicted, gt)

        self.assertTrue("recall" in results)

    def test_recall_1(self):

        predicted = np.array(
            [[10, 10, 10, 9, 9, 9], [30, 30, 30, 9, 9, 9], [50, 50, 50, 9, 9, 9]]
        )

        gt = copy.deepcopy(predicted)

        eval = RecallPrecisionMetric3D()
        results = eval.eval(predicted, gt)

        self.assertEqual(1.0, results["recall"])

    def test_recall_pred_none(self):

        predicted = None

        gt = np.array(
            [[10, 10, 10, 9, 9, 9], [30, 30, 30, 9, 9, 9], [50, 50, 50, 9, 9, 9]]
        )

        eval = RecallPrecisionMetric3D()
        results = eval.eval(predicted, gt)

        self.assertEqual(0.0, results["recall"])
        self.assertEqual(0.0, results["precision"])

    def test_recall_gt_none(self):
        predicted = np.array(
            [[10, 10, 10, 9, 9, 9], [30, 30, 30, 9, 9, 9], [50, 50, 50, 9, 9, 9]]
        )

        gt = None

        eval = RecallPrecisionMetric3D()
        results = eval.eval(predicted, gt)

        self.assertEqual(0.0, results["recall"])
        self.assertEqual(0.0, results["precision"])

    def test_precision_1(self):

        predicted = np.array(
            [[10, 10, 10, 9, 9, 9], [30, 30, 30, 9, 9, 9], [50, 50, 50, 9, 9, 9]]
        )

        gt = copy.deepcopy(predicted)

        eval = RecallPrecisionMetric3D()
        results = eval.eval(predicted, gt)

        self.assertEqual(1.0, results["precision"])

    def test_recall_2_expected_1(self):

        predicted = np.array(
            [
                [100, 10, 10, 9, 9, 9],
                [300, 30, 30, 9, 9, 9],
                [500, 50, 50, 9, 9, 9],
                [10, 10, 10, 9, 9, 9],
                [30, 30, 30, 9, 9, 9],
                [50, 50, 50, 9, 9, 9],
            ]
        )

        gt = np.array(
            [[10, 10, 10, 9, 9, 9], [30, 30, 30, 9, 9, 9], [50, 50, 50, 9, 9, 9]]
        )

        eval = RecallPrecisionMetric3D()
        results = eval.eval(predicted, gt)

        self.assertEqual(1.0, results["recall"])

    def test_recall_3_expected_2over3(self):
        predicted = np.array([[30, 30, 30, 9, 9, 9], [50, 50, 50, 9, 9, 9]])

        gt = np.array(
            [[10, 10, 10, 9, 9, 9], [30, 30, 30, 9, 9, 9], [50, 50, 50, 9, 9, 9]]
        )

        eval = RecallPrecisionMetric3D()
        results = eval.eval(predicted, gt)

        self.assertEqual(2 / 3, results["recall"])

    def test_precision_expected_05(self):

        predicted = np.array(
            [
                [100, 10, 10, 9, 9, 9],
                [300, 30, 30, 9, 9, 9],
                [500, 50, 50, 9, 9, 9],
                [10, 10, 10, 9, 9, 9],
                [30, 30, 30, 9, 9, 9],
                [50, 50, 50, 9, 9, 9],
            ]
        )

        gt = np.array(
            [[10, 10, 10, 9, 9, 9], [30, 30, 30, 9, 9, 9], [50, 50, 50, 9, 9, 9]]
        )

        eval = RecallPrecisionMetric3D()
        results = eval.eval(predicted, gt)

        self.assertEqual(0.5, results["precision"])



from mpl_toolkits.mplot3d import Axes3D
import cryolo.CoordsIO as CoordsIO
import matplotlib.pyplot as plt
import numpy

pth = "/home/twagner/Projects/Tomography/results/202009_cryolo_predict3d/more_ribosomes_michael/out_t25/CBOX_UNTRACED/d01t25_b8_lp60.cbox"
pth = "/home/twagner/Projects/Tomography/results/202009_cryolo_predict3d/pranav/out_picking_tf15_tw/CBOX_UNTRACED/dlp_fid10_049_rec.cbox"

boxes = CoordsIO.read_cbox_boxfile(pth)

xcoords = [box.x for box in boxes if box.c>0.3]
ycoords = [box.y for box in boxes if box.c>0.3]
zcoords = [box.z for box in boxes if box.c>0.3]

fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')
ax.scatter(xcoords,ycoords,zcoords)
plt.show()


import unittest
import os
import pickle
from cryolo.frontend import YOLO
import cryolo.utils as utils


class FrontendTest(unittest.TestCase):
    def test_nms_no_overlapping_boxes(self):
        from cryolo.utils import non_maxima_suppress_fast

        box = utils.BoundBox(x=10, y=10, w=10, h=10, c=0, classes=[1])
        box2 = utils.BoundBox(x=20, y=20, w=10, h=10, c=0, classes=[1])

        boxes = [box, box2]
        # front = YOLO("YOLO", [768, 768], 3, ["p"], 700, [200, 200], backend_weights=None, uniitest=True)
        nms_threshold = 0.3
        obj_threshold = 0.3
        res = non_maxima_suppress_fast(boxes, 1, nms_threshold, obj_threshold)

        self.assertEqual(len(res), 2)

    def test_nms_two_overlapping(self):
        from cryolo.utils import non_maxima_suppress_fast

        box = utils.BoundBox(x=10, y=10, w=10, h=10, c=0, classes=[1])
        box2 = utils.BoundBox(x=12, y=12, w=10, h=10, c=0, classes=[1])
        box3 = utils.BoundBox(x=20, y=20, w=10, h=10, c=0, classes=[1])
        box4 = utils.BoundBox(x=22, y=22, w=10, h=10, c=0, classes=[1])

        boxes = [box, box2, box3, box4]
        nms_threshold = 0.3
        obj_threshold = 0.3
        res = non_maxima_suppress_fast(boxes, 1, nms_threshold, obj_threshold)

        self.assertEqual(len(res), 2)

    def test_nms_two_overlapping_3d(self):
        from cryolo.utils import non_maxima_suppress_fast_3d

        boxsize = 40
        box = utils.BoundBox(
            x=100, y=100, z=100, depth=boxsize, w=boxsize, h=boxsize, c=0, classes=[1]
        )
        box2 = utils.BoundBox(
            x=109, y=109, z=109, depth=boxsize, w=boxsize, h=boxsize, c=0, classes=[1]
        )
        box3 = utils.BoundBox(
            x=200, y=200, z=200, depth=boxsize, w=boxsize, h=boxsize, c=0, classes=[1]
        )
        box4 = utils.BoundBox(
            x=201, y=201, z=201, depth=boxsize, w=boxsize, h=boxsize, c=0, classes=[1]
        )
        box.meta["num_boxes"] = 1
        box2.meta["num_boxes"] = 1
        box3.meta["num_boxes"] = 1
        box4.meta["num_boxes"] = 1
        boxes = [box, box2, box3, box4]
        nms_threshold = 0.3
        obj_threshold = 0.3
        res = non_maxima_suppress_fast_3d(boxes, nms_threshold, obj_threshold)

        self.assertEqual(len(res), 2)

    def test_nms_no_overlapping_boxes_3d(self):
        from cryolo.utils import non_maxima_suppress_fast_3d

        box = utils.BoundBox(x=10, y=10, z=10, depth=10, w=10, h=10, c=0, classes=[1])
        box2 = utils.BoundBox(x=20, y=20, z=20, depth=10, w=10, h=10, c=0, classes=[1])
        box.meta["num_boxes"] = 1
        box2.meta["num_boxes"] = 1
        boxes = [box, box2]
        # front = YOLO("YOLO", [768, 768], 3, ["p"], 700, [200, 200], backend_weights=None, uniitest=True)
        nms_threshold = 0.3
        obj_threshold = 0.3
        res = non_maxima_suppress_fast_3d(boxes, nms_threshold, obj_threshold)

        self.assertEqual(len(res), 2)


if __name__ == "__main__":
    unittest.main()



import unittest
from cryolo.evaluation.tomoevaluation import evaluate, create_parser, EvaluationResult, group_evaluation_results, run
from cryolo.evaluation.pathsboxdataprovider import PathsDataProvider
from cryolo.evaluation.pathsfilamentprovider import PathsFilamentProvider
from cryolo.evaluation.recallprecisionmetric3d import RecallPrecisionMetric3D
from cryolo.evaluation.boxdataprovider import BoxDataProvider
from cryolo.evaluation.resultsview import SimpleView
from cryolo.CoordsIO import read_cbox_boxfile
import os
import glob


class MyTestCase(unittest.TestCase):

    def test_run(self):
        path_gt = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                "../../resources/test_FolderDataProvider/01_gt_filament/",
            )
        )

        path_measured = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                "../../resources/test_FolderDataProvider/01_gt_filament/",
            )
        )
        parser = create_parser()
        args = parser.parse_args(
            ["-gtb", path_gt, "-pb", path_measured, "--filament"]
        )

        results = run(args=args, metrics=[RecallPrecisionMetric3D()], view=SimpleView())

        recall = [m.value for m in results if m.metric == "recall"][0]
        precision = [m.value for m in results if m.metric == "precision"][0]

        self.assertNotEqual(None,results)
        self.assertEqual(recall, 1.0)
        self.assertEqual(precision, 1.0)



    def test_group_evaluation_result(self):
        results = [
            EvaluationResult(
                metric="recall",
                value=0.5,
                number_datapoints=1,
                dataset="A"
            ),
            EvaluationResult(
                metric="recall",
                value=1,
                number_datapoints=1,
                dataset="A"
            ),
            EvaluationResult(
                metric="recall",
                value=1,
                number_datapoints=1,
                dataset="B"
            ),
            EvaluationResult(
                metric="recall",
                value=1,
                number_datapoints=1,
                dataset="B"
            ),
        ]

        results_grouped = group_evaluation_results(results)
        results_recall = [e.value for e in results_grouped]
        results_recall.sort()

        self.assertEqual(len(results_grouped),2)
        self.assertEqual(results_recall[0], 0.75)
        self.assertEqual(results_recall[1], 1)
        self.assertEqual(results_grouped[0].number_datapoints, 2)


    def test_evaluation_filament(self):
        path_gt = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                "../../resources/test_FolderDataProvider/01_gt_filament/*.cbox",
            )
        )

        gtfiles = glob.glob(path_gt)
        mfiles = glob.glob(path_gt)

        # As measured and gt point to the same files, recall and precision should be 1
        gtProvider = PathsFilamentProvider(
            paths=gtfiles,
            reading_method=read_cbox_boxfile,
            group_method=BoxDataProvider.group_method_lastfolder,
        )
        mProvider = PathsFilamentProvider(
            paths=mfiles,
            reading_method=read_cbox_boxfile,
            group_method=BoxDataProvider.group_method_lastfolder,
        )

        # Setup metrics
        metrics = []
        metrics.append(RecallPrecisionMetric3D())

        # Evaluate all metric on the box pairs
        result = evaluate(gtProvider, mProvider, metrics)
        recall = [m.value for m in result if m.metric == "recall"][0]
        precision = [m.value for m in result if m.metric == "precision"][0]

        self.assertEqual(recall, 1.0)
        self.assertEqual(precision, 1.0)

    def test_evaluation_particles(self):
        path_gt = os.path.join(
            os.path.dirname(__file__),
            "../../resources/test_FolderDataProvider/00_gt/*.cbox",
        )
        """
        path_measured = os.path.join(os.path.dirname(__file__),
                                     "../../resources/test_FolderDataProvider/00_measured/*.cbox")
        """
        gtfiles = glob.glob(path_gt)
        mfiles = glob.glob(path_gt)

        # As measured and gt point to the same files, recall and precision should be 1
        gtProvider = PathsDataProvider(
            paths=gtfiles,
            reading_method=read_cbox_boxfile,
            group_method=BoxDataProvider.group_method_lastfolder,
        )
        mProvider = PathsDataProvider(
            paths=mfiles,
            reading_method=read_cbox_boxfile,
            group_method=BoxDataProvider.group_method_lastfolder,
        )

        # Setup metrics
        metrics = []
        metrics.append(RecallPrecisionMetric3D())

        # Evaluate all metric on the box pairs
        result = evaluate(gtProvider, mProvider, metrics)
        recall = [m.value for m in result if m.metric == "recall"][0]
        precision = [m.value for m in result if m.metric == "precision"][0]
        self.assertEqual(recall, 1.0)
        self.assertEqual(precision, 1.0)

    def test_parser(self):
        parser = create_parser()

        gt_path = "bla/blub/"
        predicted_path = "bla/blub2/"

        args = parser.parse_args(
            ["-gtb", gt_path, "-pb", predicted_path, "--filament", "-bd", "10"]
        )

        self.assertEqual(args.groundtruth, gt_path)
        self.assertEqual(args.predicted, predicted_path)
        self.assertEqual(args.boxdistance, 10)
        self.assertEqual(args.filament, True)


if __name__ == "__main__":
    unittest.main()



from setuptools import setup
import os
import re
import codecs
import sys
import setuptools
from distutils.command.sdist import sdist
from setuptools.command.install import install
# Create new package with python setup.py sdist



here = os.path.abspath(os.path.dirname(__file__))


class PostInstallCommand(install):
    """Post-installation for installation mode."""

    def binaries_directory(self):
        import os, sys, site
        import cryolo.config_tools as _
        """Return the installation directory, or None"""
        if '--user' in sys.argv:
            paths = (site.getusersitepackages(),)
        else:
            py_version = '%s.%s' % (sys.version_info[0], sys.version_info[1])
            paths = (s % (py_version) for s in (
                sys.prefix + '/lib/python%s/dist-packages/',
                sys.prefix + '/lib/python%s/site-packages/',
                sys.prefix + '/local/lib/python%s/dist-packages/',
                sys.prefix + '/local/lib/python%s/site-packages/',
                '/Library/Python/%s/site-packages/',
            ))

        for path in paths:
            if os.path.exists(path):
                return path
        print('no installation path found', file=sys.stderr)
        return None

    def run(self):
        bin_dir = self.binaries_directory()

        init_weights = (
            "https://ndownloader.figshare.com/files/16406843"
        )  # ftp://ftp.gwdg.de/pub/misc/sphire/crYOLO-INIT/full_yolo_backend.h5'
        if bin_dir is not None:
            dir_script = os.path.join(bin_dir, "cryolo")
            init_weights_md5 = "647974a82106d7a0e5663a29a78a4598"
            FULL_YOLO_BACKEND_PATH = os.path.join(dir_script, "full_yolo_backend.h5")
            import shutil
            import urllib.request as request
            import hashlib

            from contextlib import closing
            if not os.path.exists(dir_script):
                os.makedirs(dir_script)

            if not os.path.exists(FULL_YOLO_BACKEND_PATH):
                try:
                    with closing(request.urlopen(init_weights)) as r:

                        with open(FULL_YOLO_BACKEND_PATH, "wb") as f:
                            shutil.copyfileobj(r, f)
                        print("Wrote model for initilization to:", FULL_YOLO_BACKEND_PATH)
                except Exception as e:
                    print(e)
                    print("############################################")
                    print("Unable to download initialization weights from " + init_weights)
                    print(
                        "Try to download it manually from "
                        + "https://figshare.com/articles/Initialization_weights_for_crYOLO/8965541"
                    )
                    print("and save it to " + dir_script)
                    print("############################################")
            md5sum = hashlib.md5(open(FULL_YOLO_BACKEND_PATH, "rb").read()).hexdigest()
            if md5sum != init_weights_md5:
                print("Initial weights seems to be corrupted (md5sum comparision failed).")
                if os.path.exists(FULL_YOLO_BACKEND_PATH):
                    print("Remove corrupted weights")
                    os.remove(FULL_YOLO_BACKEND_PATH)
        else:
            print("############################################")
            print("Unable to download initialization weights from " + init_weights)
            print(
                "Try to download it manually from "
                + "https://figshare.com/articles/Initialization_weights_for_crYOLO/8965541"
            )
            print("and in the cryolo train config set to the 'backend_weights' option to this file.")
            print("############################################")
        install.run(self)


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()

def find_version(*file_paths):
    print("==========================================",sys.argv)
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='cryolo',
    #version=find_version("cryolo", "__init__.py"),
    use_scm_version={'write_to': 'cryolo/__init__.py'},
    setup_requires=["Cython",'setuptools_scm'],
    python_requires='>=3.5.0, <3.9',
    packages=['cryolo','cryolo.augmentation','cryolo.evaluation','tests'],
    url='https://cryolo.readthedocs.io/en/stable/index.html',
    license='Other/Proprietary License (all rights reserved)',
    author='Thorsten Wagner',
    include_package_data=True,
    cmdclass={
        'install':PostInstallCommand
    },
    package_data={'cryolo': ['../icons/config_icon.png','../icons/program_icon.ico','../icons/program_icon.png']},

    extras_require={
        'gpu':  ['tensorflow-gpu == 1.15.4','janni[gpu] >= 0.3.0b5'],
        'cpu': ['tensorflow == 1.15.4', 'janni[cpu] >= 0.3.0b5'],
        'c11':  ['nvidia-tensorflow[horovod] == 1.15.5+nv22.2','janni >= 0.3.0b5'], #1.15.5+nv22.1
    },
    install_requires=[
        "mrcfile >= 1.3.0",
        "Cython",
        "Keras == 2.3.1",
        "numpy >= 1.16.0, < 1.19.0",
        "h5py >= 2.5.0, < 3.0.0",
        "imageio >= 2.3.0, <=2.15.0",
        "Pillow >= 6.0.0",
        "tifffile==2020.9.3",
        "networkx",
        "scipy >= 1.3.0",
        "terminaltables",
        "lineenhancer == 1.0.9.dev2",
        "cryoloBM >= 1.5",
        "ansi2html",
        "GooeyDev >= 1.0.8b5",
        "wxPython >= 4.1.0",
        "scikit-learn==0.23.2",
        "pyStarDB==0.3.1",
        "trackpy",
        "pandas==1.1.4",
        "watchdog",
        "tqdm",
        "protobuf==3.20.0"

    ],
    author_email='thorsten.wagner@mpi-dortmund.mpg.de',
    description='Picking procedure for cryo em single particle analysis',
    long_description=long_description,
    long_description_content_type='text/markdown',
    entry_points={
        'console_scripts': [
            'cryolo_train.py = cryolo.train:_main_',
            'cryolo_predict.py = cryolo.predict:_main_',
            'cryolo_evaluation.py = cryolo.eval:_main_',
            'cryolo_evaluation_tomo.py = cryolo.evaluation.tomoevaluation:_main_',
            'cryolo_gui.py = cryolo.cryolo_main:_main_']},
    classifiers=['Development Status :: 5 - Production/Stable',
                 'Environment :: Console',
                 'Environment :: X11 Applications',
                 'Intended Audience :: End Users/Desktop',
                 'Topic :: Scientific/Engineering :: Artificial Intelligence',
                 'Topic :: Scientific/Engineering :: Bio-Informatics',
                 'Programming Language :: Python :: 3',
                 'License :: Other/Proprietary License'],
)



from abc import ABC, abstractmethod
from cryolo.evaluation.tomoevaluation import EvaluationResult
from typing import List

class ResultsView(ABC):

    @abstractmethod
    def run(self, results : List[EvaluationResult]):

        pass

class SimpleView(ResultsView):

    def run(self, results: List[EvaluationResult]):

        dataset_ids = set([r.dataset for r in results])

        for datid in dataset_ids:
            results_for_one_dataset = [r for r in results if r.dataset == datid]
            print("Dataset:", datid)
            # TODO This should also take care of thresholds
            for result in results_for_one_dataset:
                print("\t", result.metric, result.value)

